"""This module contains the code to format output on the command line."""
import os
import sys
from typing import List

from rich.console import Console
from rich.status import Status
from rich.tree import Tree


_IS_WSL = "IS_WSL" in os.environ or "WSL_DISTRO_NAME" in os.environ
_IS_WINDOWS_TERMINAL = "WT_SESSION" in os.environ
_IS_WINDOWS = sys.platform == "win32"


if (_IS_WINDOWS or _IS_WSL) and not _IS_WINDOWS_TERMINAL:
    _IS_LEGACY_WINDOWS = True
else:
    _IS_LEGACY_WINDOWS = False


_COLOR_SYSTEM = None if _IS_LEGACY_WINDOWS else "auto"


ARROW_DOWN_ICON = "|" if _IS_LEGACY_WINDOWS else "⬇"
FILE_ICON = "" if _IS_LEGACY_WINDOWS else "📄 "
PYTHON_ICON = "" if _IS_LEGACY_WINDOWS else "🐍 "
TASK_ICON = "" if _IS_LEGACY_WINDOWS else "📝 "


console = Console(color_system=_COLOR_SYSTEM)


def format_strings_as_flat_tree(strings: List[str], title: str, icon: str) -> str:
    """Format list of strings as flat tree."""
    tree = Tree(title)
    for name in strings:
        tree.add(icon + name)

    text = "".join(
        [x.text for x in tree.__rich_console__(console, console.options)][:-1]
    )

    return text


def escape_squared_brackets(string: str) -> str:
    """Escape squared brackets which would be accidentally parsed by rich.

    An example are the ids of parametrized tasks which are suffixed with squared
    brackets surrounding string representations of the parametrized arguments.

    Example
    -------
    >>> escape_squared_brackets("Hello!")
    'Hello!'
    >>> escape_squared_brackets("task_dummy[arg1-arg2]")
    'task_dummy\\\\[arg1-arg2]'

    """
    return string.replace("[", "\\[")


def generate_collection_status(n_collected_tasks):
    """Generate the status object to display the progress during collection."""
    return Status(
        f"Collected {n_collected_tasks} tasks.", refresh_per_second=4, spinner="dots"
    )



"""This module contains the implementation of ``pytask collect``."""
import sys

import click
from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.console import FILE_ICON
from _pytask.console import PYTHON_ICON
from _pytask.console import TASK_ICON
from _pytask.enums import ColorCode
from _pytask.enums import ExitCode
from _pytask.exceptions import CollectionError
from _pytask.exceptions import ConfigurationError
from _pytask.nodes import reduce_node_name
from _pytask.path import find_common_ancestor
from _pytask.path import relative_to
from _pytask.pluginmanager import get_plugin_manager
from _pytask.session import Session
from rich.tree import Tree


@hookimpl(tryfirst=True)
def pytask_extend_command_line_interface(cli: click.Group):
    """Extend the command line interface."""
    cli.add_command(collect)


@hookimpl
def pytask_parse_config(config, config_from_cli):
    """Parse configuration."""
    config["nodes"] = config_from_cli.get("nodes", False)


@click.command()
@click.option("--nodes", is_flag=True, help="Show a task's dependencies and products.")
def collect(**config_from_cli):
    """Collect tasks from paths."""
    config_from_cli["command"] = "collect"

    try:
        # Duplication of the same mechanism in :func:`pytask.main.main`.
        pm = get_plugin_manager()
        from _pytask import cli

        pm.register(cli)
        pm.hook.pytask_add_hooks(pm=pm)

        config = pm.hook.pytask_configure(pm=pm, config_from_cli=config_from_cli)
        session = Session.from_config(config)

    except (ConfigurationError, Exception):
        session = Session({}, None)
        session.exit_code = ExitCode.CONFIGURATION_FAILED
        console.print_exception()

    else:
        try:
            session.hook.pytask_log_session_header(session=session)
            session.hook.pytask_collect(session=session)

            common_ancestor = _find_common_ancestor_of_all_nodes(
                session.tasks, session.config["paths"]
            )
            dictionary = _organize_tasks(session.tasks, common_ancestor)
            if dictionary:
                _print_collected_tasks(dictionary, session.config["nodes"])

            console.print()
            console.rule(style=ColorCode.NEUTRAL)

        except CollectionError:
            session.exit_code = ExitCode.COLLECTION_FAILED

        except Exception:
            session.exit_code = ExitCode.FAILED
            console.print_exception()
            console.rule(style=ColorCode.FAILED)

    sys.exit(session.exit_code)


def _find_common_ancestor_of_all_nodes(tasks, paths):
    """Find common ancestor from all nodes and passed paths."""
    all_paths = []
    for task in tasks:
        all_paths += [task.path] + [
            node.path
            for attr in ("depends_on", "produces")
            for node in getattr(task, attr).values()
        ]

    common_ancestor = find_common_ancestor(*all_paths, *paths)

    return common_ancestor


def _organize_tasks(tasks, common_ancestor):
    """Organize tasks in a dictionary.

    The dictionary has file names as keys and then a dictionary with task names and
    below a dictionary with dependencies and targets.

    """
    dictionary = {}
    for task in tasks:
        reduced_task_path = relative_to(task.path, common_ancestor)
        reduced_task_name = reduce_node_name(task, [common_ancestor])
        dictionary[reduced_task_path] = dictionary.get(reduced_task_path, {})

        task_dict = {
            reduced_task_name: {
                "depends_on": sorted(
                    relative_to(node.path, common_ancestor)
                    for node in task.depends_on.values()
                ),
                "produces": sorted(
                    relative_to(node.path, common_ancestor)
                    for node in task.produces.values()
                ),
            }
        }

        dictionary[reduced_task_path].update(task_dict)

    dictionary = {key: dictionary[key] for key in sorted(dictionary)}

    return dictionary


def _print_collected_tasks(dictionary, show_nodes):
    """Print the information on collected tasks.

    Parameters
    ----------
    dictionary: dict
        A dictionary with path on the first level, tasks on the second, dependencies and
        products on the third.
    show_nodes: bool
        Indicator for whether dependencies and products should be displayed.

    """
    # Have a new line between the number of collected tasks and this info.
    console.print()

    tree = Tree("Collected tasks:", highlight=True)
    for path in dictionary:
        module_branch = tree.add(PYTHON_ICON + f"<Module {path}>")
        for task in dictionary[path]:
            task_branch = module_branch.add(TASK_ICON + f"<Function {task}>")
            if show_nodes:
                for dependency in dictionary[path][task]["depends_on"]:
                    task_branch.add(FILE_ICON + f"<Dependency {dependency}>")

                for product in dictionary[path][task]["produces"]:
                    task_branch.add(FILE_ICON + f"<Product {product}>")

    console.print(tree)



"""This module contains the code to profile the execution."""
import csv
import json
import sys
import time
from pathlib import Path

import click
from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.database import db
from _pytask.enums import ColorCode
from _pytask.enums import ExitCode
from _pytask.exceptions import CollectionError
from _pytask.exceptions import ConfigurationError
from _pytask.nodes import FilePathNode
from _pytask.nodes import reduce_node_name
from _pytask.pluginmanager import get_plugin_manager
from _pytask.session import Session
from _pytask.shared import get_first_non_none_value
from pony import orm
from rich.table import Table
from rich.traceback import Traceback


class Runtime(db.Entity):
    """Record of runtimes of tasks."""

    task = orm.PrimaryKey(str)
    date = orm.Required(float)
    duration = orm.Required(float)


@hookimpl(tryfirst=True)
def pytask_extend_command_line_interface(cli: click.Group):
    """Extend the command line interface."""
    cli.add_command(profile)


@hookimpl
def pytask_parse_config(config, config_from_cli):
    """Parse the configuration."""
    config["export"] = get_first_non_none_value(
        config_from_cli, key="export", default=None
    )


@hookimpl
def pytask_post_parse(config):
    """Register the export option."""
    config["pm"].register(ExportNameSpace)
    config["pm"].register(DurationNameSpace)
    config["pm"].register(FileSizeNameSpace)


@hookimpl(hookwrapper=True)
def pytask_execute_task(task):
    """Attach the duration of the execution to the task."""
    start = time.time()
    yield
    end = time.time()
    task.attributes["duration"] = (start, end)


@hookimpl
def pytask_execute_task_process_report(report):
    """Store runtime of successfully finishing tasks in database."""
    task = report.task
    duration = task.attributes.get("duration")
    if report.success and duration is not None:
        _create_or_update_runtime(task.name, *duration)


@orm.db_session
def _create_or_update_runtime(task_name, start, end):
    """Create or update a runtime entry."""
    try:
        runtime = Runtime[task_name]
    except orm.ObjectNotFound:
        Runtime(task=task_name, date=start, duration=end - start)
    else:
        for attr, val in [("date", start), ("duration", end - start)]:
            setattr(runtime, attr, val)


@click.command()
@click.option(
    "--export",
    type=str,
    default=None,
    help="Export the profile in the specified format.",
)
def profile(**config_from_cli):
    """Show profile information on collected tasks."""
    config_from_cli["command"] = "profile"

    try:
        # Duplication of the same mechanism in :func:`pytask.main.main`.
        pm = get_plugin_manager()
        from _pytask import cli

        pm.register(cli)
        pm.hook.pytask_add_hooks(pm=pm)

        config = pm.hook.pytask_configure(pm=pm, config_from_cli=config_from_cli)
        session = Session.from_config(config)

    except (ConfigurationError, Exception):
        session = Session({}, None)
        session.exit_code = ExitCode.CONFIGURATION_FAILED
        console.print(Traceback.from_exception(*sys.exc_info()))

    else:
        try:
            session.hook.pytask_log_session_header(session=session)
            session.hook.pytask_collect(session=session)
            session.hook.pytask_resolve_dependencies(session=session)

            profile = {task.name: {} for task in session.tasks}
            session.hook.pytask_profile_add_info_on_task(
                session=session, tasks=session.tasks, profile=profile
            )
            profile = _process_profile(profile)

            _print_profile_table(profile, session.tasks, session.config["paths"])

            session.hook.pytask_profile_export_profile(session=session, profile=profile)

            console.rule(style=ColorCode.NEUTRAL)

        except CollectionError:
            session.exit_code = ExitCode.COLLECTION_FAILED

        except Exception:
            session.exit_code = ExitCode.FAILED
            console.print_exception()
            console.rule(style=ColorCode.FAILED)

    sys.exit(session.exit_code)


def _print_profile_table(profile, tasks, paths):
    """Print the profile table."""
    name_to_task = {task.name: task for task in tasks}
    info_names = _get_info_names(profile)

    console.print()
    if profile:
        table = Table("Task")
        for name in info_names:
            table.add_column(name, justify="right")

        for task_name, info in profile.items():
            reduced_name = reduce_node_name(name_to_task[task_name], paths)
            infos = [str(i) for i in info.values()]
            table.add_row(reduced_name, *infos)

        console.print(table)
    else:
        console.print("No information is stored on the collected tasks.")


class DurationNameSpace:
    @staticmethod
    @hookimpl
    def pytask_profile_add_info_on_task(tasks, profile):
        runtimes = _collect_runtimes([task.name for task in tasks])
        for name, duration in runtimes.items():
            profile[name]["Last Duration (in s)"] = round(duration, 2)


@orm.db_session
def _collect_runtimes(task_names):
    """Collect runtimes."""
    runtimes = [Runtime.get(task=task_name) for task_name in task_names]
    runtimes = [r for r in runtimes if r is not None]
    return {r.task: r.duration for r in runtimes}


class FileSizeNameSpace:
    @staticmethod
    @hookimpl
    def pytask_profile_add_info_on_task(session, tasks, profile):
        for task in tasks:
            successors = list(session.dag.successors(task.name))
            if successors:
                sum_bytes = 0
                for successor in successors:
                    node = session.dag.nodes[successor]["node"]
                    if isinstance(node, FilePathNode):
                        try:
                            sum_bytes += node.path.stat().st_size
                        except FileNotFoundError:
                            pass

                profile[task.name]["Size of Products"] = _to_human_readable_size(
                    sum_bytes
                )


def _to_human_readable_size(bytes_, units=None):
    """Convert bytes to a human readable size."""
    units = [" bytes", "KB", "MB", "GB", "TB"] if units is None else units
    return (
        str(bytes_) + units[0]
        if bytes_ < 1024
        else _to_human_readable_size(bytes_ >> 10, units[1:])
    )


def _process_profile(profile):
    """Process profile to make it ready for printing and storing."""
    info_names = _get_info_names(profile)
    if info_names:
        complete_profiles = {
            task_name: {
                attr_name: profile[task_name].get(attr_name, "")
                for attr_name in info_names
            }
            for task_name in sorted(profile)
        }
    else:
        complete_profiles = {}
    return complete_profiles


class ExportNameSpace:
    @staticmethod
    @hookimpl(trylast=True)
    def pytask_profile_export_profile(session, profile):
        extension = session.config["export"]

        if extension == "csv":
            _export_to_csv(profile)
        elif extension == "json":
            _export_to_json(profile)
        elif extension is None:
            pass
        else:
            raise ValueError(f"The export option '{extension}' cannot be handled.")


def _export_to_csv(profile):
    """Export profile to csv."""
    info_names = _get_info_names(profile)
    path = Path.cwd().joinpath("profile.csv")

    with open(path, "w", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(("Task", *info_names))
        for task_name, info in profile.items():
            writer.writerow((task_name, *info.values()))


def _export_to_json(profile):
    """Export profile to json."""
    json_ = json.dumps(profile)
    path = Path.cwd().joinpath("profile.json")
    path.write_text(json_)


def _get_info_names(profile):
    """Get names of infos of tasks."""
    info_names = sorted(set().union(*(set(val) for val in profile.values())))
    return info_names



"""Implement the database managed with pony."""
from pathlib import Path

import click
from _pytask.config import hookimpl
from _pytask.shared import convert_truthy_or_falsy_to_bool
from _pytask.shared import get_first_non_none_value
from pony import orm


db = orm.Database()


class State(db.Entity):
    """Represent the state of a node in relation to a task."""

    task = orm.Required(str)
    node = orm.Required(str)
    state = orm.Required(str)

    orm.PrimaryKey(task, node)


def create_database(provider, filename, create_db, create_tables):
    """Create the database.

    Raises
    ------
    orm.BindingError
        Raise if database is already bound.

    """
    try:
        db.bind(provider=provider, filename=filename, create_db=create_db)
        db.generate_mapping(create_tables=create_tables)
    except orm.BindingError:
        pass


@orm.db_session
def create_or_update_state(first_key, second_key, state):
    """Create or update a state."""
    try:
        state_in_db = State[first_key, second_key]
    except orm.ObjectNotFound:
        State(task=first_key, node=second_key, state=state)
    else:
        state_in_db.state = state


@hookimpl
def pytask_extend_command_line_interface(cli):
    """Extend command line interface."""
    additional_parameters = [
        click.Option(
            ["--database-provider"],
            type=click.Choice(["sqlite", "postgres", "mysql", "oracle", "cockroach"]),
            help=(
                "Database provider. All providers except sqlite are considered "
                "experimental.  [default: sqlite]"
            ),
            default=None,
        ),
        click.Option(
            ["--database-filename"],
            type=click.Path(),
            help="Path to database relative to root.  [default: .pytask.sqlite3]",
            default=None,
        ),
        click.Option(
            ["--database-create-db"],
            type=bool,
            help="Create database if it does not exist.  [default: True]",
            default=None,
        ),
        click.Option(
            ["--database-create-tables"],
            type=bool,
            help="Create tables if they do not exist.  [default: True]",
            default=None,
        ),
    ]
    cli.commands["build"].params.extend(additional_parameters)


@hookimpl
def pytask_parse_config(config, config_from_cli, config_from_file):
    """Parse the configuration."""
    config["database_provider"] = get_first_non_none_value(
        config_from_cli, config_from_file, key="database_provider", default="sqlite"
    )
    filename = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="database_filename",
        default=".pytask.sqlite3",
    )
    filename = Path(filename)
    if not filename.is_absolute():
        filename = Path(config["root"], filename).resolve()
    config["database_filename"] = filename

    config["database_create_db"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="database_create_db",
        default=True,
        callback=convert_truthy_or_falsy_to_bool,
    )
    config["database_create_tables"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="database_create_tables",
        default=True,
        callback=convert_truthy_or_falsy_to_bool,
    )
    config["database"] = {
        "provider": config["database_provider"],
        "filename": config["database_filename"].as_posix(),
        "create_db": config["database_create_db"],
        "create_tables": config["database_create_tables"],
    }


@hookimpl
def pytask_post_parse(config):
    create_database(**config["database"])



import copy
import functools
import itertools
import pprint
import types
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from _pytask.config import hookimpl
from _pytask.console import format_strings_as_flat_tree
from _pytask.console import TASK_ICON
from _pytask.mark import Mark
from _pytask.mark import MARK_GEN as mark  # noqa: N811
from _pytask.shared import find_duplicates


def parametrize(
    arg_names: Union[str, Tuple[str], List[str]],
    arg_values: Iterable,
    *,
    ids: Optional[
        Union[Iterable[Union[bool, float, int, str, None]], Callable[[Any], Any]]
    ] = None,
):
    """Parametrize a task function.

    Parametrizing a task allows to execute the same task with different arguments.

    Parameters
    ----------
    arg_names : Union[str, Tuple[str], List[str]]
        The names of the arguments which can either be given as a comma-separated
        string, a tuple of strings, or a list of strings.
    arg_values : Iterable
        The values which correspond to names in ``arg_names``. For one argument, it is a
        single iterable. For multiple argument names it is an iterable of iterables.
    ids
        This argument can either be a list with ids or a function which is called with
        every value passed to the parametrized function.

        If you pass an iterable with ids, make sure to only use :obj:`bool`,
        :obj:`float`, :obj:`int`, or :obj:`str` as values which are used to create task
        ids like ``"task_dummpy.py::task_dummy[first_task_id]"``.

        If you pass a function, the function receives each value of the parametrization
        and may return a boolean, number, string or None. For the latter, the
        auto-generated value is used.

    """
    return arg_names, arg_values, ids


@hookimpl
def pytask_parse_config(config):
    config["markers"]["parametrize"] = (
        "Call a task function multiple times passing in different arguments each turn. "
        "arg_values generally needs to be a list of values if arg_names specifies only "
        "one name or a list of tuples of values if arg_names specifies multiple "
        "names.Example: @pytask.mark.parametrize('arg1', [1, 2]) would lead to two "
        "calls of the decorated task function, one with arg1=1 and another with arg1=2."
        " See this [link https://bit.ly/3vqyiAk]tutorial (https://bit.ly/3vqyiAk)[/] "
        "for more info and examples."
    )


@hookimpl
def pytask_parametrize_task(session, name, obj):
    """Parametrize a task.

    This function takes a single Python function and all parametrize decorators and
    generates multiple instances of the same task with different arguments.

    Note that, while a single ``@pytask.mark.parametrize`` is handled like a loop or a
    :func:`zip`, multiple ``@pytask.mark.parametrize`` decorators form a Cartesian
    product.

    We cannot raise an error if the function does not use parametrized arguments since
    some plugins will replace functions with their own implementation like pytask-r.

    """
    if callable(obj):
        obj, markers = _remove_parametrize_markers_from_func(obj)

        if len(markers) > 1:
            raise NotImplementedError(
                "Multiple parametrizations are currently not implemented since it is "
                "not possible to define products for tasks from a cartesian product."
            )

        base_arg_names, arg_names, arg_values = _parse_parametrize_markers(
            markers, name
        )

        product_arg_names = list(itertools.product(*arg_names))
        product_arg_values = list(itertools.product(*arg_values))

        names_and_functions = []
        for names, values in zip(product_arg_names, product_arg_values):
            kwargs = dict(
                zip(
                    itertools.chain.from_iterable(base_arg_names),
                    itertools.chain.from_iterable(values),
                )
            )

            # Copy function and attributes to allow in-place changes.
            func = _copy_func(obj)
            func.pytaskmark = copy.deepcopy(obj.pytaskmark)

            # Convert parametrized dependencies and products to decorator.
            session.hook.pytask_parametrize_kwarg_to_marker(obj=func, kwargs=kwargs)

            # Attach remaining parametrized arguments to the function.
            partialed_func = functools.partial(func, **kwargs)
            wrapped_func = functools.update_wrapper(partialed_func, func)

            name_ = f"{name}[{'-'.join(itertools.chain.from_iterable(names))}]"
            names_and_functions.append((name_, wrapped_func))

        names = [i[0] for i in names_and_functions]
        duplicates = find_duplicates(names)
        if duplicates:
            text = format_strings_as_flat_tree(
                duplicates, "Duplicated task ids", TASK_ICON
            )
            raise ValueError(
                "The following ids are duplicated while parametrizing task "
                f"'{obj.__name__}'.\n\n{text}\n\nIt might be caused by "
                "parametrizing the task with the same combination of arguments "
                "multiple times. Change the arguments or change the ids generated by "
                "the parametrization."
            )

        return names_and_functions


def _remove_parametrize_markers_from_func(obj):
    """Remove parametrize markers from the object."""
    parametrize_markers = [i for i in obj.pytaskmark if i.name == "parametrize"]
    others = [i for i in obj.pytaskmark if i.name != "parametrize"]
    obj.pytaskmark = others

    return obj, parametrize_markers


def _parse_parametrize_marker(marker: Mark, name: str) -> Tuple[Any, Any, Any]:
    """Parse parametrize marker.

    Parameters
    ----------
    marker : Mark
        A parametrize mark.
    name : str
        The name of the task function which is parametrized.

    Returns
    -------
    base_arg_names : Tuple[str]
        Contains the names of the arguments.
    processed_arg_names : List[Tuple[str]]
        Each tuple in the list represents the processed names of the arguments suffixed
        with a number indicating the iteration.
    processed_arg_values : List[Tuple[Any]]
        Each tuple in the list represents the values of the arguments for each
        iteration.

    """
    arg_names, arg_values, ids = parametrize(*marker.args, **marker.kwargs)

    parsed_arg_names = _parse_arg_names(arg_names)
    parsed_arg_values = _parse_arg_values(arg_values)

    _check_if_n_arg_names_matches_n_arg_values(
        parsed_arg_names, parsed_arg_values, name
    )

    expanded_arg_names = _create_parametrize_ids_components(
        parsed_arg_names, parsed_arg_values, ids
    )

    return parsed_arg_names, expanded_arg_names, parsed_arg_values


def _parse_parametrize_markers(markers: List[Mark], name: str):
    """Parse parametrize markers."""
    parsed_markers = [_parse_parametrize_marker(marker, name) for marker in markers]
    base_arg_names = [i[0] for i in parsed_markers]
    processed_arg_names = [i[1] for i in parsed_markers]
    processed_arg_values = [i[2] for i in parsed_markers]

    return base_arg_names, processed_arg_names, processed_arg_values


def _parse_arg_names(arg_names: Union[str, Tuple[str], List[str]]) -> Tuple[str]:
    """Parse arg_names argument of parametrize decorator.

    There are three allowed formats:

    1. comma-separated string representation.
    2. a tuple of strings.
    3. a list of strings.

    All formats are converted to a tuple of strings.

    Parameters
    ----------
    arg_names : Union[str, Tuple[str], List[str]]
        The names of the arguments which are parametrized.

    Returns
    -------
    out : Tuple[str]
        The parsed arg_names.

    Example
    -------
    >>> _parse_arg_names("i")
    ('i',)
    >>> _parse_arg_names("i, j")
    ('i', 'j')

    """
    if isinstance(arg_names, str):
        out = tuple(i.strip() for i in arg_names.split(","))
    elif isinstance(arg_names, (tuple, list)):
        out = tuple(arg_names)
    else:
        raise ValueError(
            "The argument 'arg_names' accepts comma-separated strings, tuples and lists"
            f" of strings. It cannot accept {arg_names} with type {type(arg_names)}."
        )

    return out


def _parse_arg_values(arg_values: Iterable) -> List[Tuple[Any]]:
    """Parse the values provided for each argument name.

    Example
    -------
    >>> _parse_arg_values(["a", "b", "c"])
    [('a',), ('b',), ('c',)]
    >>> _parse_arg_values([(0, 0), (0, 1), (1, 0)])
    [(0, 0), (0, 1), (1, 0)]

    """
    return [
        tuple(i) if isinstance(i, Iterable) and not isinstance(i, str) else (i,)
        for i in arg_values
    ]


def _check_if_n_arg_names_matches_n_arg_values(
    arg_names: Tuple[str], arg_values: List[Tuple[Any]], name: str
) -> None:
    """Check if the number of argument names matches the number of arguments."""
    n_names = len(arg_names)
    n_values = [len(i) for i in arg_values]
    unique_n_values = tuple(set(n_values))

    if not all(i == n_names for i in unique_n_values):
        pretty_arg_values = (
            f"{unique_n_values[0]}"
            if len(unique_n_values) == 1
            else " or ".join(map(str, unique_n_values))
        )
        idx_example = [i == n_names for i in n_values].index(False)
        formatted_example = pprint.pformat(arg_values[idx_example])
        raise ValueError(
            f"Task '{name}' is parametrized with {n_names} 'arg_names', {arg_names}, "
            f"but the number of provided 'arg_values' is {pretty_arg_values}. For "
            f"example, here are the values of parametrization no. {idx_example}:"
            f"\n\n{formatted_example}"
        )


def _create_parametrize_ids_components(
    arg_names: Tuple[str],
    arg_values: List[Tuple[Any]],
    ids: Optional[
        Union[Iterable[Union[bool, float, int, str, None]], Callable[[Any], Any]]
    ],
):
    """Create the ids for each parametrization.

    Parameters
    ----------
    arg_names : Tuple[str]
        The names of the arguments of the parametrized function.
    arg_values : List[Tuple[Any]]
        A list of tuples where each tuple is for one run.
    ids
        The ids associated with one parametrization.

    Examples
    --------
    >>> _create_parametrize_ids_components(["i"], [(0,), (1,)], None)
    [('0',), ('1',)]

    >>> _create_parametrize_ids_components(["i", "j"], [(0, (0,)), (1, (1,))], None)
    [('0', 'j0'), ('1', 'j1')]

    """
    if isinstance(ids, Iterable):
        parsed_ids = [(id_,) for id_ in ids]

        if len(parsed_ids) != len(arg_values):
            raise ValueError("The number of ids must match the number of runs.")

        if not all(
            isinstance(id_, (bool, int, float, str)) or id_ is None
            for id_ in itertools.chain.from_iterable(parsed_ids)
        ):
            raise ValueError(
                "Ids for parametrization can only be of type bool, float, int, str or "
                "None."
            )

        parsed_ids = [(str(id_),) for id_ in itertools.chain.from_iterable(parsed_ids)]

    else:
        parsed_ids = []
        for i, _arg_values in enumerate(arg_values):
            id_components = tuple(
                _arg_value_to_id_component(arg_names[j], arg_value, i, ids)
                for j, arg_value in enumerate(_arg_values)
            )
            parsed_ids.append(id_components)

    return parsed_ids


def _arg_value_to_id_component(
    arg_name: str, arg_value: Any, i: int, id_func: Union[Callable[[Any], Any], None]
) -> str:
    """Create id component from the name and value of the argument.

    First, transform the value of the argument with a user-defined function if given.
    Otherwise, take the original value. Then, if the value is a :obj:`bool`,
    :obj:`float`, :obj:`int`, or :obj:`str`, cast it to a string. Otherwise, define a
    placeholder value from the name of the argument and the iteration.

    Parameters
    ----------
    arg_name : str
        Name of the parametrized function argument.
    arg_value : Any
        Value of the argument.
    i : int
        The ith iteration of the parametrization.
    id_func : Union[Callable[[Any], Any], None]
        A callable which maps argument values to :obj:`bool`, :obj:`float`, :obj:`int`,
        or :obj:`str` or anything else. Any object with a different dtype than the first
        will be mapped to an auto-generated id component.

    Returns
    -------
    id_component : str
        A part of the final parametrized id.

    """
    id_component = id_func(arg_value) if id_func is not None else None
    if isinstance(id_component, (bool, float, int, str)):
        id_component = str(id_component)
    elif isinstance(arg_value, (bool, float, int, str)):
        id_component = str(arg_value)
    else:
        id_component = arg_name + str(i)
    return id_component


@hookimpl
def pytask_parametrize_kwarg_to_marker(obj, kwargs: dict) -> None:
    """Add some parametrized keyword arguments as decorator."""
    if callable(obj):
        for marker_name in ["depends_on", "produces"]:
            if marker_name in kwargs:
                mark.__getattr__(marker_name)(kwargs.pop(marker_name))(obj)


def _copy_func(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """Create a copy of a function.

    Based on https://stackoverflow.com/a/13503277/7523785.

    Example
    -------
    >>> def _func(): pass
    >>> copied_func = _copy_func(_func)
    >>> _func is copied_func
    False

    """
    new_func = types.FunctionType(
        func.__code__,
        func.__globals__,
        name=func.__name__,
        argdefs=func.__defaults__,
        closure=func.__closure__,
    )
    new_func = functools.update_wrapper(new_func, func)
    new_func.__kwdefaults__ = func.__kwdefaults__
    return new_func



import sys
from typing import AbstractSet

import attr
import click
from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.enums import ExitCode
from _pytask.exceptions import ConfigurationError
from _pytask.mark.expression import Expression
from _pytask.mark.expression import ParseError
from _pytask.mark.structures import get_marks_from_obj
from _pytask.mark.structures import get_specific_markers_from_task
from _pytask.mark.structures import has_marker
from _pytask.mark.structures import Mark
from _pytask.mark.structures import MARK_GEN
from _pytask.mark.structures import MarkDecorator
from _pytask.mark.structures import MarkGenerator
from _pytask.pluginmanager import get_plugin_manager
from _pytask.session import Session
from _pytask.shared import convert_truthy_or_falsy_to_bool
from _pytask.shared import get_first_non_none_value
from rich.table import Table


__all__ = [
    "get_specific_markers_from_task",
    "get_marks_from_obj",
    "has_marker",
    "Expression",
    "Mark",
    "MarkDecorator",
    "MarkGenerator",
    "MARK_GEN",
    "ParseError",
]


@click.command()
def markers(**config_from_cli):
    """Show all registered markers."""
    config_from_cli["command"] = "markers"

    try:
        # Duplication of the same mechanism in :func:`pytask.main.main`.
        pm = get_plugin_manager()
        from _pytask import cli

        pm.register(cli)
        pm.hook.pytask_add_hooks(pm=pm)

        config = pm.hook.pytask_configure(pm=pm, config_from_cli=config_from_cli)
        session = Session.from_config(config)

    except (ConfigurationError, Exception):
        console.print_exception()
        session = Session({}, None)
        session.exit_code = ExitCode.CONFIGURATION_FAILED

    else:
        table = Table("Marker", "Description", leading=1)

        for name, description in config["markers"].items():
            table.add_row(f"pytask.mark.{name}", description)

        console.print(table)

    sys.exit(session.exit_code)


@hookimpl
def pytask_extend_command_line_interface(cli: click.Group) -> None:
    """Add marker related options."""
    additional_build_parameters = [
        click.Option(
            ["--strict-markers"],
            is_flag=True,
            help="Raise errors for unknown markers.",
            default=None,
        ),
        click.Option(
            ["-m", "marker_expression"],
            metavar="MARKER_EXPRESSION",
            type=str,
            help="Select tasks via marker expressions.",
        ),
        click.Option(
            ["-k", "expression"],
            metavar="EXPRESSION",
            type=str,
            help="Select tasks via expressions on task ids.",
        ),
    ]
    cli.commands["build"].params.extend(additional_build_parameters)
    cli.commands["clean"].params.extend(additional_build_parameters)
    cli.commands["collect"].params.extend(additional_build_parameters)

    cli.add_command(markers)


@hookimpl
def pytask_parse_config(config, config_from_cli, config_from_file):
    """Parse marker related options."""
    markers = _read_marker_mapping_from_ini(config_from_file.get("markers", ""))
    config["markers"] = {**markers, **config["markers"]}
    config["strict_markers"] = get_first_non_none_value(
        config,
        config_from_file,
        config_from_cli,
        key="strict_markers",
        default=False,
        callback=convert_truthy_or_falsy_to_bool,
    )

    config["expression"] = config_from_cli.get("expression")
    config["marker_expression"] = config_from_cli.get("marker_expression")

    MARK_GEN.config = config


def _read_marker_mapping_from_ini(string: str) -> dict:
    """Read marker descriptions from configuration file."""
    # Split by newlines and remove empty strings.
    lines = filter(lambda x: bool(x), string.split("\n"))
    mapping = {}
    for line in lines:
        try:
            key, value = line.split(":")
        except ValueError as e:
            key = line
            value = ""
            if not key.isidentifier():
                raise ValueError(
                    f"{key} is not a valid Python name and cannot be used as a marker."
                ) from e

        mapping[key] = value

    return mapping


@attr.s(slots=True)
class KeywordMatcher:
    """A matcher for keywords.

    Given a list of names, matches any substring of one of these names. The string
    inclusion check is case-insensitive.

    Will match on the name of the task, including the names of its parents. Only matches
    names of items which are either a :class:`Class` or :class:`Function`.

    Additionally, matches on names in the 'extra_keyword_matches' set of any task, as
    well as names directly assigned to test functions.

    """

    _names = attr.ib(type=AbstractSet[str])

    @classmethod
    def from_task(cls, task) -> "KeywordMatcher":
        mapped_names = {task.name}

        # Add the names attached to the current function through direct assignment.
        function_obj = task.function
        if function_obj:
            mapped_names.update(function_obj.__dict__)

        # Add the markers to the keywords as we no longer handle them correctly.
        mapped_names.update(mark.name for mark in task.markers)

        return cls(mapped_names)

    def __call__(self, subname: str) -> bool:
        subname = subname.lower()
        names = (name.lower() for name in self._names)

        for name in names:
            if subname in name:
                return True
        return False


def deselect_by_keyword(session, tasks) -> None:
    """Deselect tests by keywords."""
    keywordexpr = session.config["expression"]
    if not keywordexpr:
        return

    try:
        expression = Expression.compile_(keywordexpr)
    except ParseError as e:
        raise ValueError(
            f"Wrong expression passed to '-k': {keywordexpr}: {e}"
        ) from None

    remaining = []
    deselected = []
    for task in tasks:
        if keywordexpr and not expression.evaluate(KeywordMatcher.from_task(task)):
            deselected.append(task)
        else:
            remaining.append(task)

    if deselected:
        session.deselected.extend(deselected)
        tasks[:] = remaining


@attr.s(slots=True)
class MarkMatcher:
    """A matcher for markers which are present.

    Tries to match on any marker names, attached to the given colitem.

    """

    own_mark_names = attr.ib()

    @classmethod
    def from_task(cls, task) -> "MarkMatcher":
        mark_names = {mark.name for mark in task.markers}
        return cls(mark_names)

    def __call__(self, name: str) -> bool:
        return name in self.own_mark_names


def deselect_by_mark(session, tasks) -> None:
    """Deselect tests by marks."""
    matchexpr = session.config["marker_expression"]
    if not matchexpr:
        return

    try:
        expression = Expression.compile_(matchexpr)
    except ParseError as e:
        raise ValueError(f"Wrong expression passed to '-m': {matchexpr}: {e}") from None

    remaining = []
    deselected = []
    for task in tasks:
        if expression.evaluate(MarkMatcher.from_task(task)):
            remaining.append(task)
        else:
            deselected.append(task)

    if deselected:
        session.deselected.extend(deselected)
        tasks[:] = remaining


@hookimpl
def pytask_collect_modify_tasks(session, tasks):
    """Modify the list of collected tasks with expressions and markers."""
    deselect_by_keyword(session, tasks)
    deselect_by_mark(session, tasks)



"""Configure pytask."""
import configparser
import itertools
import os
import tempfile
import warnings
from pathlib import Path
from typing import List

import pluggy
from _pytask.shared import convert_truthy_or_falsy_to_bool
from _pytask.shared import get_first_non_none_value
from _pytask.shared import parse_paths
from _pytask.shared import parse_value_or_multiline_option
from _pytask.shared import to_list


hookimpl = pluggy.HookimplMarker("pytask")


_IGNORED_FOLDERS = [
    ".git/*",
    ".hg/*",
    ".svn/*",
    ".venv/*",
]


_IGNORED_FILES = [
    ".codecov.yml",
    ".gitignore",
    ".pre-commit-config.yaml",
    ".readthedocs.yml",
    ".readthedocs.yaml",
    "readthedocs.yml",
    "readthedocs.yaml",
    "environment.yml",
    "pytask.ini",
    "setup.cfg",
    "tox.ini",
]


_IGNORED_FILES_AND_FOLDERS = _IGNORED_FILES + _IGNORED_FOLDERS


IGNORED_TEMPORARY_FILES_AND_FOLDERS = [
    "*.egg-info/*",
    ".ipynb_checkpoints/*",
    ".mypy_cache/*",
    ".nox/*",
    ".tox/*",
    "_build/*",
    "__pycache__/*",
    "build/*",
    "dist/*",
    "pytest_cache/*",
]


def is_file_system_case_sensitive() -> bool:
    """Check whether the file system is case-sensitive."""
    with tempfile.NamedTemporaryFile(prefix="TmP") as tmp_file:
        return not os.path.exists(tmp_file.name.lower())


IS_FILE_SYSTEM_CASE_SENSITIVE = is_file_system_case_sensitive()


@hookimpl
def pytask_configure(pm, config_from_cli):
    """Configure pytask."""
    config = {"pm": pm}

    # Either the path to the configuration is passed via the CLI or it needs to be
    # detected from the paths passed to pytask.
    if config_from_cli.get("config"):
        config["config"] = Path.cwd().joinpath(config_from_cli["config"])
        config["root"] = config["config"].parent
    else:
        paths = (
            parse_paths(config_from_cli.get("paths"))
            if config_from_cli.get("paths") is not None
            else [Path.cwd()]
        )
        config["root"], config["config"] = _find_project_root_and_ini(paths)

    config_from_file = (
        _read_config(config["config"]) if config["config"] is not None else {}
    )

    # If paths are set in the configuration, process them.
    if config_from_file.get("paths"):
        paths_from_file = parse_value_or_multiline_option(config_from_file.get("paths"))
        config_from_file["paths"] = [
            config["config"].parent.joinpath(p).resolve() for p in paths_from_file
        ]

    config["paths"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="paths",
        default=[Path.cwd()],
        callback=parse_paths,
    )

    config["markers"] = {
        "depends_on": "Attach a dependency/dependencies to a task.",
        "produces": "Attach a product/products to a task.",
        "try_first": "Try to execute a task a early as possible.",
        "try_last": "Try to execute a task a late as possible.",
    }

    pm.hook.pytask_parse_config(
        config=config,
        config_from_cli=config_from_cli,
        config_from_file=config_from_file,
    )

    pm.hook.pytask_post_parse(config=config)

    return config


@hookimpl
def pytask_parse_config(config, config_from_cli, config_from_file):
    """Parse the configuration."""
    config["command"] = config_from_cli.get("command", "build")

    config_from_file["ignore"] = parse_value_or_multiline_option(
        config_from_file.get("ignore")
    )
    config["ignore"] = (
        to_list(
            get_first_non_none_value(
                config_from_cli,
                config_from_file,
                key="ignore",
                default=[],
            )
        )
        + _IGNORED_FILES_AND_FOLDERS
        + IGNORED_TEMPORARY_FILES_AND_FOLDERS
    )

    config["debug_pytask"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="debug_pytask",
        default=False,
        callback=convert_truthy_or_falsy_to_bool,
    )
    if config["debug_pytask"]:
        config["pm"].trace.root.setwriter(print)  # noqa: T002
        config["pm"].enable_tracing()

    config_from_file["task_files"] = parse_value_or_multiline_option(
        config_from_file.get("task_files")
    )
    config["task_files"] = to_list(
        get_first_non_none_value(
            config_from_file, key="task_files", default="task_*.py"
        )
    )

    config["stop_after_first_failure"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="stop_after_first_failure",
        default=False,
        callback=convert_truthy_or_falsy_to_bool,
    )
    if config["stop_after_first_failure"]:
        config["max_failures"] = 1
    else:
        config["max_failures"] = get_first_non_none_value(
            config_from_cli,
            config_from_file,
            key="max_failures",
            default=float("inf"),
            callback=lambda x: x if x is None else int(x),
        )

    config["check_casing_of_paths"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="check_casing_of_paths",
        default=True,
        callback=convert_truthy_or_falsy_to_bool,
    )


@hookimpl
def pytask_post_parse(config):
    """Sort markers alphabetically."""
    config["markers"] = {k: config["markers"][k] for k in sorted(config["markers"])}


def _find_project_root_and_ini(paths: List[Path]):
    """Find the project root and configuration file from a list of paths."""
    try:
        common_ancestor = Path(os.path.commonpath(paths))
    except ValueError:
        warnings.warn(
            "A common path for all passed path could not be detected. Fall back to "
            "current working directory."
        )
        common_ancestor = Path.cwd()

    config_path = None
    parent_directories = (
        common_ancestor.parents
        if common_ancestor.is_file()
        else [common_ancestor] + list(common_ancestor.parents)
    )
    for parent, config_name in itertools.product(
        parent_directories, ["pytask.ini", "tox.ini", "setup.cfg"]
    ):
        path = parent.joinpath(config_name)

        if path.exists():
            try:
                _read_config(path)
            except Exception:
                pass
            else:
                config_path = path
                break

    if config_path is not None:
        root = config_path.parent
    else:
        root = common_ancestor if common_ancestor.is_dir() else common_ancestor.parent

    return root, config_path


def _read_config(path):
    """Read the configuration from a file with a [pytask] section."""
    config = configparser.ConfigParser()
    config.read(path)
    return dict(config["pytask"])



import warnings
from typing import Any
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import attr


def get_specific_markers_from_task(task, marker_name):
    return [marker for marker in task.markers if marker.name == marker_name]


def get_marks_from_obj(obj, marker_name):
    return [
        marker
        for marker in getattr(obj, "pytaskmark", [])
        if marker.name == marker_name
    ]


def has_marker(obj, marker_name):
    return any(marker.name == marker_name for marker in getattr(obj, "pytaskmark", []))


def is_task_function(func) -> bool:
    return callable(func) and getattr(func, "__name__", "<lambda>") != "<lambda>"


@attr.s(frozen=True)
class Mark:
    name = attr.ib(type=str)
    """str: Name of the mark."""
    args = attr.ib(type=Tuple[Any, ...])
    """Tuple[Any]: Positional arguments of the mark decorator."""
    kwargs = attr.ib(type=Mapping[str, Any])
    """Mapping[str, Any]: Keyword arguments of the mark decorator."""

    _param_ids_from = attr.ib(type=Optional["Mark"], default=None, repr=False)
    """Optional[Mark]: Source Mark for ids with parametrize Marks."""
    _param_ids_generated = attr.ib(
        type=Optional[Sequence[str]], default=None, repr=False
    )
    """Optional[Sequence[str]]: Resolved/generated ids with parametrize Marks."""

    def _has_param_ids(self) -> bool:
        return "ids" in self.kwargs or len(self.args) >= 4

    def combined_with(self, other: "Mark") -> "Mark":
        """Return a new Mark which is a combination of this Mark and another Mark.

        Combines by appending args and merging kwargs.

        Parameters
        ----------
        other : pytask.mark.structures.Mark
            The mark to combine with.

        Returns
        -------
        Mark
            The new mark which is a combination of two marks.

        """
        assert self.name == other.name

        # Remember source of ids with parametrize Marks.
        param_ids_from: Optional[Mark] = None
        if self.name == "parametrize":
            if other._has_param_ids():
                param_ids_from = other
            elif self._has_param_ids():
                param_ids_from = self

        return Mark(
            self.name,
            self.args + other.args,
            dict(self.kwargs, **other.kwargs),
            param_ids_from=param_ids_from,
        )


@attr.s
class MarkDecorator:
    """A decorator for applying a mark on task function.

    MarkDecorators are created with ``pytask.mark``

    .. code-block:: python

        mark1 = pytask.mark.NAME  # Simple MarkDecorator
        mark2 = pytask.mark.NAME(name1=value)  # Parametrized MarkDecorator

    and can then be applied as decorators to task functions

    .. code-block:: python

        @mark2
        def task_function():
            pass

    When a MarkDecorator is called it does the following:

    1. If called with a single function as its only positional argument and no
       additional keyword arguments, it attaches the mark to the function, containing
       all the arguments already stored internally in the MarkDecorator.
    2. When called in any other case, it returns a new MarkDecorator instance with the
       original MarkDecorator's content updated with the arguments passed to this call.

    Notes
    -----
    The rules above prevent MarkDecorators from storing only a single function or class
    reference as their positional argument with no additional keyword or positional
    arguments. You can work around this by using `with_args()`.

    """

    mark = attr.ib(type=Mark, validator=attr.validators.instance_of(Mark))

    @property
    def name(self) -> str:
        """Alias for mark.name."""
        return self.mark.name

    @property
    def args(self) -> Tuple[Any, ...]:
        """Alias for mark.args."""
        return self.mark.args

    @property
    def kwargs(self) -> Mapping[str, Any]:
        """Alias for mark.kwargs."""
        return self.mark.kwargs

    def __repr__(self) -> str:
        return f"<MarkDecorator {self.mark!r}>"

    def with_args(self, *args: object, **kwargs: object) -> "MarkDecorator":
        """Return a MarkDecorator with extra arguments added.

        Unlike calling the MarkDecorator, ``with_args()`` can be used even if the sole
        argument is a callable.

        """
        mark = Mark(self.name, args, kwargs)
        return self.__class__(self.mark.combined_with(mark))

    def __call__(self, *args: object, **kwargs: object):  # noqa: F811
        """Call the MarkDecorator."""
        if args and not kwargs:
            func = args[0]
            if len(args) == 1 and is_task_function(func):
                store_mark(func, self.mark)
                return func
        return self.with_args(*args, **kwargs)


def get_unpacked_marks(obj) -> List[Mark]:
    """Obtain the unpacked marks that are stored on an object."""
    mark_list = getattr(obj, "pytaskmark", [])
    if not isinstance(mark_list, list):
        mark_list = [mark_list]
    return normalize_mark_list(mark_list)


def normalize_mark_list(mark_list: Iterable[Union[Mark, MarkDecorator]]) -> List[Mark]:
    """Normalizes marker decorating helpers to mark objects.

    Parameters
    ----------
    mark_list : List[Union[Mark, MarkDecorator]]

    Returns
    -------
    List[Mark]

    """
    extracted = [
        getattr(mark, "mark", mark) for mark in mark_list
    ]  # unpack MarkDecorator
    for mark in extracted:
        if not isinstance(mark, Mark):
            raise TypeError(f"got {mark!r} instead of Mark")
    return [x for x in extracted if isinstance(x, Mark)]


def store_mark(obj, mark: Mark) -> None:
    """Store a Mark on an object.
    This is used to implement the Mark declarations/decorators correctly.
    """
    assert isinstance(mark, Mark), mark
    # Always reassign name to avoid updating pytaskmark in a reference that
    # was only borrowed.
    obj.pytaskmark = get_unpacked_marks(obj) + [mark]


class MarkGenerator:
    """Factory for :class:`MarkDecorator` objects - exposed as a ``pytask.mark``
    singleton instance.

    Example
    -------
    >>> import pytask

    >>> @pytask.mark.skip
    ... def task_function():
    ...    pass

    applies a 'skip' :class:`Mark` on ``task_function``.

    """

    config = None
    """Optional[dict]: The configuration."""
    markers = set()
    """Set[str]: The set of markers."""

    def __getattr__(self, name: str) -> MarkDecorator:
        if name[0] == "_":
            raise AttributeError("Marker name must NOT start with underscore")

        if self.config is not None:
            # We store a set of markers as a performance optimization - if a mark
            # name is in the set we definitely know it, but a mark may be known and
            # not in the set.  We therefore start by updating the set!
            if name not in self.markers:
                self.markers.update(self.config["markers"])

            # If the name is not in the set of known marks after updating,
            # then it really is time to issue a warning or an error.
            if name not in self.markers:
                if self.config["strict_markers"]:
                    raise ValueError(f"Unknown pytask.mark.{name}.")
                # Raise a specific error for common misspellings of "parametrize".
                if name in ["parameterize", "parametrise", "parameterise"]:
                    warnings.warn(f"Unknown '{name}' mark, did you mean 'parametrize'?")

                warnings.warn(
                    f"Unknown pytask.mark.{name} - is this a typo? You can register "
                    "custom marks to avoid this warning.",
                    stacklevel=2,
                )

        return MarkDecorator(Mark(name, (), {}))


MARK_GEN = MarkGenerator()



"""Implement hook specifications or entry-points for pytask.

At each of the entry-points, a plugin can register a hook implementation which receives
the message send by the host and may send a response.

"""
import click
import pluggy


hookspec = pluggy.HookspecMarker("pytask")


@hookspec
def pytask_add_hooks(pm: pluggy.PluginManager) -> None:
    """Add hook specifications and implementations to the plugin manager.

    This hook is the first to be called to let plugins register their hook
    specifications and implementations.

    If you want to register plugins dynamically depending on the configuration, use
    :func:`pytask_post_parse` instead. See :mod:`_pytask.debugging` for an example.

    """


# Hooks for the command-line interface.


@hookspec
def pytask_extend_command_line_interface(cli: click.Group) -> None:
    """Extend the command line interface.

    The hook can be used to extend the command line interface either by providing new
    commands or adding options and arguments to existing commands.

    - Add commands via ``cli.add_command``.
    - Add options and arguments by extending ``cli.params`` or
      ``cli.commands["<name>"].params`` if the parameters only apply to a certain
      command.

    """


# Hooks for the configuration.


@hookspec(firstresult=True)
def pytask_configure(pm: pluggy.PluginManager, config_from_cli: dict) -> dict:
    """Configure pytask.

    The main hook implementation which controls the configuration and calls subordinated
    hooks.

    """


@hookspec
def pytask_parse_config(
    config: dict, config_from_cli: dict, config_from_file: dict
) -> None:
    """Parse configuration from the CLI or from file.

    This hook can be used to unify the configuration from the command line interface,
    the configuration file and provided defaults. The function
    :func:`pytask.shared.get_first_non_none_value` might be helpful for that.

    Note that, the configuration is changed in-place.

    """


@hookspec
def pytask_post_parse(config: dict) -> None:
    """Post parsing.

    This hook allows to consolidate the configuration in case some plugins might be
    mutually exclusive. For example, the parallel execution provided by pytask-parallel
    does not work with any form of debugging. If debugging is turned on, parallelization
    can be turned of in this step.

    """


@hookspec
def pytask_unconfigure(session):
    """Unconfigure a pytask session before the process is exited.

    The hook allows to return resources previously borrowed like :func:`pdb.set_trace`
    by :class:`_pytask.debugging.PytaskPDB` and do other stuff at the end of a session.

    """


# Hooks for the collection.


@hookspec(firstresult=True)
def pytask_collect(session):
    """Collect tasks from paths.

    The main hook implementation which controls the collection and calls subordinated
    hooks.

    """


@hookspec(firstresult=True)
def pytask_ignore_collect(path, config):
    """Ignore collected path.

    This hook is indicates for each directory and file whether it should be ignored.
    This speeds up the collection.

    """


@hookspec
def pytask_collect_modify_tasks(session, tasks):
    """Modify tasks after they have been collected.

    This hook can be used to deselect tasks when they match a certain keyword or mark.

    """


@hookspec(firstresult=True)
def pytask_collect_file_protocol(session, path, reports):
    """Start protocol to collect files.

    The protocol calls the subordinate hook :func:`pytask_collect_file` which might
    error if the file has a :class:`SyntaxError`.

    """


@hookspec(firstresult=True)
def pytask_collect_file(session, path, reports):
    """Collect tasks from a file.

    If you want to collect tasks from other files, modify this hook.

    """


@hookspec(firstresult=True)
def pytask_collect_task_protocol(session, path, name, obj):
    """Start protocol to collect tasks."""


@hookspec
def pytask_collect_task_setup(session, path, name, obj):
    """Steps before collecting a task."""


@hookspec(firstresult=True)
def pytask_collect_task(session, path, name, obj):
    """Collect a single task."""


@hookspec
def pytask_collect_task_teardown(session, task):
    """Perform tear-down operations when a task was collected.

    Use this hook specification to, for example, perform checks on the collected task.

    """


@hookspec(firstresult=True)
def pytask_collect_node(session, path, node):
    """Collect a node which is a dependency or a product of a task."""


@hookspec(firstresult=True)
def pytask_collect_log(session, reports, tasks):
    """Log errors occurring during the collection.

    This hook reports errors during the collection.

    """


# Hooks to parametrize tasks.


@hookspec(firstresult=True)
def pytask_parametrize_task(session, name, obj):
    """Generate multiple tasks from name and object with parametrization."""


@hookspec
def pytask_parametrize_kwarg_to_marker(obj, kwargs):
    """Add some keyword arguments as markers to object.

    This hook moves arguments defined in the parametrization to marks of the same
    function. This allows an argument like ``depends_on`` be transformed to the usual
    ``@pytask.mark.depends_on`` marker which receives special treatment.

    """


# Hooks for resolving dependencies.


@hookspec(firstresult=True)
def pytask_resolve_dependencies(session):
    """Resolve dependencies.

    The main hook implementation which controls the resolution of dependencies and calls
    subordinated hooks.

    """


@hookspec(firstresult=True)
def pytask_resolve_dependencies_create_dag(session, tasks):
    """Create the DAG.

    This hook creates the DAG from tasks, dependencies and products. The DAG can be used
    by a scheduler to find an execution order.

    """


@hookspec(firstresult=True)
def pytask_resolve_dependencies_select_execution_dag(session, dag):
    """Select the subgraph which needs to be executed.

    This hook determines which of the tasks have to be re-run because something has
    changed.

    """


@hookspec(firstresult=True)
def pytask_resolve_dependencies_validate_dag(session, dag):
    """Validate the DAG.

    This hook validates the DAG. For example, there can be cycles in the DAG if tasks,
    dependencies and products have been misspecified.

    """


@hookspec
def pytask_resolve_dependencies_log(session, report):
    """Log errors during resolving dependencies."""


# Hooks for running tasks.


@hookspec(firstresult=True)
def pytask_execute(session):
    """Loop over all tasks for the execution.

    The main hook implementation which controls the execution and calls subordinated
    hooks.

    """


@hookspec
def pytask_execute_log_start(session):
    """Start logging of execution.

    This hook allows to provide a header with information before the execution starts.

    """


@hookspec(firstresult=True)
def pytask_execute_create_scheduler(session):
    """Create a scheduler for the execution.

    The scheduler provides information on which tasks are able to be executed. Its
    foundation is likely a topological ordering of the tasks based on the DAG.

    """


@hookspec(firstresult=True)
def pytask_execute_build(session):
    """Execute the build.

    This hook implements the main loop to execute tasks.

    """


@hookspec(firstresult=True)
def pytask_execute_task_protocol(session, task):
    """Run the protocol for executing a test.

    This hook runs all stages of the execution process, setup, execution, and teardown
    and catches any exception.

    Then, the exception or success is stored in a report and logged.

    """


@hookspec
def pytask_execute_task_log_start(session, task):
    """Start logging of task execution.

    This hook can be used to provide more verbose output during the execution.

    """


@hookspec
def pytask_execute_task_setup(session, task):
    """Set up the task execution.

    This hook is called before the task is executed and can provide an entry-point to
    fast-fail a task. For example, raise and exception if a dependency is missing
    instead of letting the error occur in the execution.

    """


@hookspec(firstresult=True)
def pytask_execute_task(session, task):
    """Execute a task."""


@hookspec
def pytask_execute_task_teardown(session, task):
    """Tear down task execution.

    This hook is executed after the task has been executed. It allows to perform
    clean-up operations or checks for missing products.

    """


@hookspec(firstresult=True)
def pytask_execute_task_process_report(session, report):
    """Process the report of a task.

    This hook allows to process each report generated by a task which is either based on
    an exception or a success. Set the color and the symbol for logging.

    Some exceptions are intentionally raised like skips, but they should not be reported
    as failures.

    """


@hookspec(firstresult=True)
def pytask_execute_task_log_end(session, task, report):
    """Log the end of a task execution."""


@hookspec
def pytask_execute_log_end(session, reports):
    """Log the footer of the execution report."""


# Hooks for general logging.


@hookspec
def pytask_log_session_header(session):
    """Log session information at the begin of a run."""


@hookspec
def pytask_log_session_footer(session, infos, duration, color):
    """Log session information at the end of a run."""


# Hooks for profile.


@hookspec
def pytask_profile_add_info_on_task(session, tasks, profile):
    """Add information on task for profile.

    Hook implementations can add information to the ``profile`` dictionary. The
    dictionary's keys are the task names. The value for each task is a dictionary itself
    where keys correspond to columns of the profile table.

    """


@hookspec
def pytask_profile_export_profile(session, profile):
    """Export the profile."""



"""This file contains the command and code for drawing the DAG."""
import shutil
from pathlib import Path
from typing import Any
from typing import Dict

import click
import networkx as nx
from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.dag import descending_tasks
from _pytask.enums import ColorCode
from _pytask.enums import ExitCode
from _pytask.exceptions import CollectionError
from _pytask.exceptions import ConfigurationError
from _pytask.exceptions import ResolvingDependenciesError
from _pytask.nodes import reduce_names_of_multiple_nodes
from _pytask.pluginmanager import get_plugin_manager
from _pytask.session import Session
from _pytask.shared import get_first_non_none_value


@hookimpl(tryfirst=True)
def pytask_extend_command_line_interface(cli: click.Group):
    """Extend the command line interface."""
    cli.add_command(dag)


@hookimpl
def pytask_parse_config(config, config_from_cli, config_from_file):
    """Parse configuration."""
    config["output_path"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="output_path",
        default=Path.cwd() / "dag.pdf",
        callback=lambda x: None if x is None else Path(x),
    )
    config["layout"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="layout",
        default="dot",
    )


_HELP_TEXT_LAYOUT = (
    "The layout determines the structure of the graph. Here you find an overview of "
    "all available layouts: https://graphviz.org/#roadmap."
)


_HELP_TEXT_OUTPUT = (
    "The output path of the visualization. The format is inferred from the file "
    "extension."
)


@click.command()
@click.option("-l", "--layout", type=str, default=None, help=_HELP_TEXT_LAYOUT)
@click.option("-o", "--output-path", type=str, default=None, help=_HELP_TEXT_OUTPUT)
def dag(**config_from_cli):
    """Create a visualization of the project's DAG."""
    session = _create_session(config_from_cli)
    dag = _refine_dag(session)
    _write_graph(dag, session.config["output_path"], session.config["layout"])


def build_dag(config_from_cli: Dict[str, Any]) -> "pydot.Dot":  # noqa: F821
    """Build the DAG.

    This function is the programmatic interface to ``pytask dag`` and returns a
    preprocessed :class:`pydot.Dot` which makes plotting easier than with matplotlib.

    To change the style of the graph, it might be easier to convert the graph back to
    networkx, set attributes, and convert back to pydot or pygraphviz.

    Parameters
    ----------
    config_from_cli : Dict[str, Any]
        The configuration usually received from the CLI. For example, use ``{"paths":
        "example-directory/"}`` to collect tasks from a directory.

    Returns
    -------
    pydot.Dot
        A preprocessed graph which can be customized and exported.

    """
    session = _create_session(config_from_cli)
    dag = _refine_dag(session)
    return dag


def _refine_dag(session):
    dag = _shorten_node_labels(session.dag, session.config["paths"])
    dag = _add_root_node(dag)
    dag = _clean_dag(dag)
    dag = _style_dag(dag)
    dag = _escape_node_names_with_colons(dag)

    return dag


def _create_session(config_from_cli: Dict[str, Any]) -> nx.DiGraph:
    try:
        pm = get_plugin_manager()
        from _pytask import cli

        pm.register(cli)
        pm.hook.pytask_add_hooks(pm=pm)

        config = pm.hook.pytask_configure(pm=pm, config_from_cli=config_from_cli)

        session = Session.from_config(config)

    except (ConfigurationError, Exception):
        console.print_exception()
        session = Session({}, None)
        session.exit_code = ExitCode.CONFIGURATION_FAILED

    else:
        try:
            session.hook.pytask_log_session_header(session=session)
            session.hook.pytask_collect(session=session)
            session.hook.pytask_resolve_dependencies(session=session)

        except CollectionError:
            session.exit_code = ExitCode.COLLECTION_FAILED

        except ResolvingDependenciesError:
            session.exit_code = ExitCode.RESOLVING_DEPENDENCIES_FAILED

        except Exception:
            session.exit_code = ExitCode.FAILED
            console.print_exception()
            console.rule(style=ColorCode.FAILED)

    return session


def _shorten_node_labels(dag, paths):
    node_names = dag.nodes
    short_names = reduce_names_of_multiple_nodes(node_names, dag, paths)
    old_to_new = dict(zip(node_names, short_names))
    dag = nx.relabel_nodes(dag, old_to_new)
    return dag


def _add_root_node(dag):
    tasks_without_predecessor = [
        name
        for name in dag.nodes
        if len(list(descending_tasks(name, dag))) == 0 and "task" in dag.nodes[name]
    ]
    if tasks_without_predecessor:
        dag.add_node("root")
        for name in tasks_without_predecessor:
            dag.add_edge("root", name)

    return dag


def _clean_dag(dag):
    """Clean the DAG."""
    for node in dag.nodes:
        dag.nodes[node].clear()
    return dag


def _style_dag(dag: nx.DiGraph) -> nx.DiGraph:
    shapes = {name: "hexagon" if "::task_" in name else "box" for name in dag.nodes}
    nx.set_node_attributes(dag, shapes, "shape")
    return dag


def _escape_node_names_with_colons(dag: nx.DiGraph):
    """Escape node names with colons.

    pydot cannot handle colons in node names since it messes up some syntax. Escaping
    works by wrapping the string in double quotes. See this issue for more information:
    https://github.com/pydot/pydot/issues/224.

    """
    return nx.relabel_nodes(dag, {name: f'"{name}"' for name in dag.nodes})


def _write_graph(dag: nx.DiGraph, path: Path, layout: str) -> None:
    try:
        import pydot  # noqa: F401
    except ImportError:
        raise ImportError(
            "To visualize the project's DAG you need to install pydot which is "
            "available with pip and conda. For example, use 'conda install -c "
            "conda-forge pydot'."
        ) from None
    if shutil.which(layout) is None:
        raise RuntimeError(
            f"The layout program '{layout}' could not be found on your PATH. Please, "
            "install graphviz. For example, use 'conda install -c conda-forge "
            "graphivz'."
        )

    path.parent.mkdir(exist_ok=True, parents=True)
    graph = nx.nx_pydot.to_pydot(dag)
    graph.write(path, prog=layout, format=path.suffix[1:])



import sys
import time

from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.dag import descending_tasks
from _pytask.dag import node_and_neighbors
from _pytask.dag import TopologicalSorter
from _pytask.database import create_or_update_state
from _pytask.enums import ColorCode
from _pytask.exceptions import ExecutionError
from _pytask.exceptions import NodeNotFoundError
from _pytask.mark import Mark
from _pytask.nodes import FilePathNode
from _pytask.nodes import reduce_node_name
from _pytask.report import ExecutionReport
from rich.traceback import Traceback


@hookimpl
def pytask_execute(session):
    """Execute tasks."""
    session.hook.pytask_execute_log_start(session=session)
    session.scheduler = session.hook.pytask_execute_create_scheduler(session=session)
    session.hook.pytask_execute_build(session=session)
    session.hook.pytask_execute_log_end(
        session=session, reports=session.execution_reports
    )


@hookimpl
def pytask_execute_log_start(session):
    """Start logging."""
    session.execution_start = time.time()

    # New line to separate note on collected items from task statuses.
    console.print()


@hookimpl(trylast=True)
def pytask_execute_create_scheduler(session):
    """Create a scheduler based on topological sorting."""
    scheduler = TopologicalSorter.from_dag(session.dag)
    scheduler.prepare()
    return scheduler


@hookimpl
def pytask_execute_build(session):
    """Execute tasks."""
    for name in session.scheduler.static_order():
        task = session.dag.nodes[name]["task"]
        report = session.hook.pytask_execute_task_protocol(session=session, task=task)
        session.execution_reports.append(report)
        if session.should_stop:
            return True

    return True


@hookimpl
def pytask_execute_task_protocol(session, task):
    """Follow the protocol to execute each task."""
    session.hook.pytask_execute_task_log_start(session=session, task=task)
    try:
        session.hook.pytask_execute_task_setup(session=session, task=task)
        session.hook.pytask_execute_task(session=session, task=task)
        session.hook.pytask_execute_task_teardown(session=session, task=task)
    except Exception:
        report = ExecutionReport.from_task_and_exception(task, sys.exc_info())
    else:
        report = ExecutionReport.from_task(task)
    session.hook.pytask_execute_task_process_report(session=session, report=report)
    session.hook.pytask_execute_task_log_end(session=session, task=task, report=report)

    return report


@hookimpl(trylast=True)
def pytask_execute_task_setup(session, task):
    """Set up the execution of a task.

    1. Check whether all dependencies of a task are available.
    2. Create the directory where the product will be placed.

    """
    for dependency in session.dag.predecessors(task.name):
        node = session.dag.nodes[dependency]["node"]
        try:
            node.state()
        except NodeNotFoundError:
            raise NodeNotFoundError(
                f"{node.name} is missing and required for {task.name}."
            )

    # Create directory for product if it does not exist. Maybe this should be a `setup`
    # method for the node classes.
    for product in session.dag.successors(task.name):
        node = session.dag.nodes[product]["node"]
        if isinstance(node, FilePathNode):
            node.path.parent.mkdir(parents=True, exist_ok=True)


@hookimpl
def pytask_execute_task(task):
    """Execute task."""
    task.execute()


@hookimpl
def pytask_execute_task_teardown(session, task):
    """Check if each produced node was indeed produced."""
    for product in session.dag.successors(task.name):
        node = session.dag.nodes[product]["node"]
        try:
            node.state()
        except NodeNotFoundError:
            raise NodeNotFoundError(f"{node.name} was not produced by {task.name}.")


@hookimpl(trylast=True)
def pytask_execute_task_process_report(session, report):
    """Process the execution report of a task.

    If a task failed, skip all subsequent tasks. Else, update the states of related
    nodes in the database.

    """
    task = report.task
    if report.success:
        _update_states_in_database(session.dag, task.name)
        report.symbol = "."
        report.color = ColorCode.SUCCESS
    else:
        report.symbol = "F"
        report.color = ColorCode.FAILED
        for descending_task_name in descending_tasks(task.name, session.dag):
            descending_task = session.dag.nodes[descending_task_name]["task"]
            descending_task.markers.append(
                Mark(
                    "skip_ancestor_failed",
                    (),
                    {"reason": f"Previous task '{task.name}' failed."},
                )
            )

        session.n_tasks_failed += 1
        if session.n_tasks_failed >= session.config["max_failures"]:
            session.should_stop = True

    return True


@hookimpl(trylast=True)
def pytask_execute_task_log_end(report):
    """Log task outcome."""
    console.print(report.symbol, style=report.color, end="")


@hookimpl
def pytask_execute_log_end(session, reports):
    session.execution_end = time.time()
    show_locals = session.config["show_locals"]

    n_successful = sum(report.success for report in reports)
    n_failed = len(reports) - n_successful

    console.print()
    any_failure = any(not report.success for report in reports)
    if any_failure:
        console.rule(f"[{ColorCode.FAILED}]Failures", style=ColorCode.FAILED)
        console.print()

    for report in reports:
        if not report.success:

            task_name = reduce_node_name(report.task, session.config["paths"])
            if len(task_name) > console.width - 15:
                task_name = report.task.base_name
            console.rule(
                f"[{ColorCode.FAILED}]Task {task_name} failed", style=ColorCode.FAILED
            )

            console.print()

            console.print(
                Traceback.from_exception(*report.exc_info, show_locals=show_locals)
            )

            console.print()
            show_capture = session.config["show_capture"]
            for when, key, content in report.sections:
                if key in ("stdout", "stderr") and show_capture in (key, "all"):
                    console.rule(f"Captured {key} during {when}", style=None)
                    console.print(content)

    session.hook.pytask_log_session_footer(
        session=session,
        infos=[
            (n_successful, "succeeded", ColorCode.SUCCESS),
            (n_failed, "failed", ColorCode.FAILED),
        ],
        duration=round(session.execution_end - session.execution_start, 2),
        color=ColorCode.FAILED if n_failed else ColorCode.SUCCESS,
    )

    if n_failed:
        raise ExecutionError

    return True


def _update_states_in_database(dag, task_name):
    """Update the state for each node of a task in the database."""
    for name in node_and_neighbors(dag, task_name):
        node = dag.nodes[name].get("task") or dag.nodes[name]["node"]
        create_or_update_state(task_name, node.name, node.state())



"""Functions which are used across various modules."""
import glob
from collections.abc import Sequence
from pathlib import Path
from typing import Any
from typing import Iterable


def to_list(scalar_or_iter):
    """Convert scalars and iterables to list.

    Parameters
    ----------
    scalar_or_iter : str or list

    Returns
    -------
    list

    Examples
    --------
    >>> to_list("a")
    ['a']
    >>> to_list(["b"])
    ['b']

    """
    return (
        [scalar_or_iter]
        if isinstance(scalar_or_iter, str) or not isinstance(scalar_or_iter, Sequence)
        else list(scalar_or_iter)
    )


def parse_paths(x):
    """Parse paths."""
    if x is not None:
        paths = [Path(p) for p in to_list(x)]
        paths = [
            Path(p).resolve() for path in paths for p in glob.glob(path.as_posix())
        ]
        return paths


def falsy_to_none_callback(ctx, param, value):  # noqa: U100
    """Convert falsy object to ``None``.

    Some click arguments accept multiple inputs and instead of ``None`` as a default if
    no information is passed, they return empty lists or tuples.

    Since pytask uses ``None`` as a placeholder value for skippable inputs, convert the
    values.

    Examples
    --------
    >>> falsy_to_none_callback(None, None, ()) is None
    True
    >>> falsy_to_none_callback(None, None, []) is None
    True
    >>> falsy_to_none_callback(None, None, 1)
    1

    """
    return value if value else None


def get_first_non_none_value(*configs, key, default=None, callback=None):
    """Get the first non-None value for a key from a list of dictionaries.

    This function allows to prioritize information from many configurations by changing
    the order of the inputs while also providing a default.

    Examples
    --------
    >>> get_first_non_none_value({"a": None}, {"a": 1}, key="a")
    1
    >>> get_first_non_none_value({"a": None}, {"a": None}, key="a", default="default")
    'default'
    >>> get_first_non_none_value({}, {}, key="a", default="default")
    'default'
    >>> get_first_non_none_value({"a": None}, {"a": "b"}, key="a")
    'b'

    """
    callback = (lambda x: x) if callback is None else callback  # noqa: E731
    processed_values = (callback(config.get(key)) for config in configs)
    return next((value for value in processed_values if value is not None), default)


def parse_value_or_multiline_option(value):
    """Parse option which can hold a single value or values separated by new lines."""
    if value in ["none", "None", None, ""]:
        value = None
    elif isinstance(value, str) and "\n" in value:
        value = [v.strip() for v in value.split("\n") if v.strip()]
    elif isinstance(value, str) and "n" not in value:
        value = value.strip()
    return value


def convert_truthy_or_falsy_to_bool(x):
    """Convert truthy or falsy value in .ini to Python boolean."""
    if x in [True, "True", "true", "1"]:
        out = True
    elif x in [False, "False", "false", "0"]:
        out = False
    elif x in [None, "None", "none"]:
        out = None
    else:
        raise ValueError(
            f"Input '{x}' is neither truthy (True, true, 1) or falsy (False, false, 0)."
        )
    return out


def find_duplicates(x: Iterable[Any]):
    """Find duplicated entries in iterable.

    Examples
    --------
    >>> find_duplicates(["a", "b", "a"])
    {'a'}
    >>> find_duplicates(["a", "b"])
    set()

    """
    seen = set()
    duplicates = set()

    for i in x:
        if i in seen:
            duplicates.add(i)
        seen.add(i)

    return duplicates



"""Implement some capabilities to deal with the DAG."""
import itertools
from pathlib import Path
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List

import attr
import networkx as nx
from _pytask.console import format_strings_as_flat_tree
from _pytask.console import TASK_ICON
from _pytask.mark import get_specific_markers_from_task
from _pytask.nodes import MetaTask
from _pytask.nodes import reduce_node_name


def descending_tasks(task_name: str, dag: nx.DiGraph) -> Generator[str, None, None]:
    """Yield only descending tasks."""
    for descendant in nx.descendants(dag, task_name):
        if "task" in dag.nodes[descendant]:
            yield descendant


def task_and_descending_tasks(
    task_name: str, dag: nx.DiGraph
) -> Generator[str, None, None]:
    """Yield task and descending tasks."""
    yield task_name
    yield from descending_tasks(task_name, dag)


def node_and_neighbors(dag: nx.DiGraph, node: str) -> Generator[str, None, None]:
    """Yield node and neighbors which are first degree predecessors and successors.

    We cannot use ``dag.neighbors`` as it only considers successors as neighbors in a
    DAG.

    """
    return itertools.chain([node], dag.predecessors(node), dag.successors(node))


@attr.s
class TopologicalSorter:
    """The topological sorter class.

    This class allows to perform a topological sort

    """

    dag = attr.ib(converter=nx.DiGraph)
    priorities = attr.ib(factory=dict)
    _dag_backup = attr.ib(default=None)
    _is_prepared = attr.ib(default=False, type=bool)
    _nodes_out = attr.ib(factory=set)

    @classmethod
    def from_dag(cls, dag: nx.DiGraph, paths: List[Path] = None) -> "TopologicalSorter":
        if paths is None:
            paths = []

        if not dag.is_directed():
            raise ValueError("Only directed graphs have a topological order.")

        tasks = [
            dag.nodes[node]["task"] for node in dag.nodes if "task" in dag.nodes[node]
        ]
        priorities = _extract_priorities_from_tasks(tasks, paths)

        task_names = {task.name for task in tasks}
        task_dict = {name: nx.ancestors(dag, name) & task_names for name in task_names}
        task_dag = nx.DiGraph(task_dict).reverse()

        return cls(task_dag, priorities, task_dag.copy())

    def prepare(self) -> None:
        """Perform some checks before creating a topological ordering."""
        try:
            nx.algorithms.cycles.find_cycle(self.dag)
        except nx.NetworkXNoCycle:
            pass
        else:
            raise ValueError("The DAG contains cycles.")

        self._is_prepared = True

    def get_ready(self, n: int = 1):
        """Get up to ``n`` tasks which are ready."""
        if not self._is_prepared:
            raise ValueError("The TopologicalSorter needs to be prepared.")
        if not isinstance(n, int) or n < 1:
            raise ValueError("'n' must be an integer greater or equal than 1.")

        ready_nodes = {v for v, d in self.dag.in_degree() if d == 0} - self._nodes_out
        prioritized_nodes = sorted(
            ready_nodes, key=lambda x: self.priorities.get(x, 0)
        )[-n:]

        self._nodes_out.update(prioritized_nodes)

        return prioritized_nodes

    def is_active(self) -> bool:
        """Indicate whether there are still tasks left."""
        return bool(self.dag.nodes)

    def done(self, *nodes: Iterable[str]) -> None:
        """Mark some tasks as done."""
        self._nodes_out = self._nodes_out - set(nodes)
        self.dag.remove_nodes_from(nodes)

    def reset(self) -> None:
        """Reset an exhausted topological sorter."""
        self.dag = self._dag_backup.copy()
        self._is_prepared = False
        self._nodes_out = set()

    def static_order(self) -> Generator[str, None, None]:
        """Return a topological order of tasks as an iterable."""
        self.prepare()
        while self.is_active():
            new_task = self.get_ready()[0]
            yield new_task
            self.done(new_task)


def _extract_priorities_from_tasks(
    tasks: List[MetaTask], paths: List[Path]
) -> Dict[str, int]:
    """Extract priorities from tasks.

    Priorities are set via the ``pytask.mark.try_first`` and ``pytask.mark.try_last``
    markers. We recode these markers to numeric values to sort all available by
    priorities. ``try_first`` is assigned the highest value such that it has the
    rightmost position in the list. Then, we can simply call :meth:`list.pop` on the
    list which is far more efficient than ``list.pop(0)``.

    """
    priorities = {
        task.name: {
            "try_first": bool(get_specific_markers_from_task(task, "try_first")),
            "try_last": bool(get_specific_markers_from_task(task, "try_last")),
        }
        for task in tasks
    }
    tasks_w_mixed_priorities = [
        name for name, p in priorities.items() if p["try_first"] and p["try_last"]
    ]

    if tasks_w_mixed_priorities:
        name_to_task = {task.name: task for task in tasks}
        reduced_names = [
            reduce_node_name(name_to_task[name], paths)
            for name in tasks_w_mixed_priorities
        ]
        text = format_strings_as_flat_tree(
            reduced_names, "Tasks with mixed priorities", TASK_ICON
        )
        raise ValueError(
            "'try_first' and 'try_last' cannot be applied on the same task. See the "
            f"following tasks for errors:\n\n{text}"
        )

    # Recode to numeric values for sorting.
    numeric_mapping = {(True, False): 1, (False, False): 0, (False, True): -1}
    numeric_priorities = {
        name: numeric_mapping[(p["try_first"], p["try_last"])]
        for name, p in priorities.items()
    }

    return numeric_priorities



"""Deals with nodes which are dependencies or products of a task."""
import functools
import inspect
import itertools
import pathlib
from abc import ABCMeta
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union

import attr
from _pytask.console import escape_squared_brackets
from _pytask.exceptions import NodeNotCollectedError
from _pytask.exceptions import NodeNotFoundError
from _pytask.mark import get_marks_from_obj
from _pytask.path import find_closest_ancestor
from _pytask.path import find_common_ancestor
from _pytask.path import relative_to
from _pytask.shared import find_duplicates


def depends_on(objects: Union[Any, Iterable[Any]]) -> Union[Any, Iterable[Any]]:
    """Specify dependencies for a task.

    Parameters
    ----------
    objects : Union[Any, Iterable[Any]]
        Can be any valid Python object or an iterable of any Python objects. To be
        valid, it must be parsed by some hook implementation for the
        :func:`pytask.hookspecs.pytask_collect_node` entry-point.

    """
    return objects


def produces(objects: Union[Any, Iterable[Any]]) -> Union[Any, Iterable[Any]]:
    """Specify products of a task.

    Parameters
    ----------
    objects : Union[Any, Iterable[Any]]
        Can be any valid Python object or an iterable of any Python objects. To be
        valid, it must be parsed by some hook implementation for the
        :func:`pytask.hookspecs.pytask_collect_node` entry-point.

    """
    return objects


class MetaNode(metaclass=ABCMeta):
    """Meta class for nodes."""

    @abstractmethod
    def state(self):
        """Return a value which indicates whether a node has changed or not."""
        pass


class MetaTask(MetaNode):
    """The base class for tasks."""

    @abstractmethod
    def execute(self):
        """Execute the task."""
        pass


@attr.s
class PythonFunctionTask(MetaTask):
    """The class for tasks which are Python functions."""

    base_name = attr.ib(type=str)
    """str: The base name of the task."""
    name = attr.ib(type=str)
    """str: The unique identifier for a task."""
    path = attr.ib(type=Path)
    """pathlib.Path: Path to the file where the task was defined."""
    function = attr.ib(type=callable)
    """callable: The task function."""
    depends_on = attr.ib(factory=dict)
    """Optional[List[MetaNode]]: A list of dependencies of task."""
    produces = attr.ib(factory=dict)
    """List[MetaNode]: A list of products of task."""
    markers = attr.ib(factory=list)
    """Optional[List[Mark]]: A list of markers attached to the task function."""
    keep_dict = attr.ib(default=False)
    """Dict[str, bool]: Should dictionaries for single nodes be preserved?"""
    _report_sections = attr.ib(factory=list)
    """List[Tuple[str]]: A list of reports with entries for when, what, and content."""
    attributes = attr.ib(factory=dict)
    """Dict[Any, Any]: A dictionary to store additional information of the task."""

    @classmethod
    def from_path_name_function_session(cls, path, name, function, session):
        """Create a task from a path, name, function, and session."""
        keep_dictionary = {}

        objects = _extract_nodes_from_function_markers(function, depends_on)
        nodes, keep_dict = _convert_objects_to_node_dictionary(objects, "depends_on")
        keep_dictionary["depends_on"] = keep_dict
        dependencies = _collect_nodes(session, path, name, nodes)

        objects = _extract_nodes_from_function_markers(function, produces)
        nodes, keep_dict = _convert_objects_to_node_dictionary(objects, "produces")
        keep_dictionary["produces"] = keep_dict
        products = _collect_nodes(session, path, name, nodes)

        markers = [
            marker
            for marker in getattr(function, "pytaskmark", [])
            if marker.name not in ["depends_on", "produces"]
        ]

        return cls(
            base_name=name,
            name=create_task_name(path, name),
            path=path,
            function=function,
            depends_on=dependencies,
            produces=products,
            markers=markers,
            keep_dict=keep_dictionary,
        )

    def execute(self):
        """Execute the task."""
        kwargs = self._get_kwargs_from_task_for_function()
        self.function(**kwargs)

    def state(self):
        """Return the last modified date of the file where the task is defined."""
        return str(self.path.stat().st_mtime)

    def _get_kwargs_from_task_for_function(self):
        """Process dependencies and products to pass them as kwargs to the function."""
        func_arg_names = set(inspect.signature(self.function).parameters)
        kwargs = {}
        for arg_name in ["depends_on", "produces"]:
            if arg_name in func_arg_names:
                attribute = getattr(self, arg_name)
                kwargs[arg_name] = (
                    attribute[0].value
                    if len(attribute) == 1
                    and 0 in attribute
                    and not self.keep_dict[arg_name]
                    else {name: node.value for name, node in attribute.items()}
                )

        return kwargs

    def add_report_section(self, when: str, key: str, content: str):
        if content:
            self._report_sections.append((when, key, content))


@attr.s
class FilePathNode(MetaNode):
    """The class for a node which is a path."""

    name = attr.ib()
    """str: Name of the node which makes it identifiable in the DAG."""

    value = attr.ib()
    """Any: Value passed to the decorator which can be requested inside the function."""

    path = attr.ib()
    """pathlib.Path: Path to the FilePathNode."""

    @classmethod
    @functools.lru_cache()
    def from_path(cls, path: pathlib.Path):
        """Instantiate class from path to file.

        The `lru_cache` decorator ensures that the same object is not collected twice.

        """
        if not path.is_absolute():
            raise ValueError("FilePathNode must be instantiated from absolute path.")
        return cls(path.as_posix(), path, path)

    def state(self):
        """Return the last modified date for file path."""
        if not self.path.exists():
            raise NodeNotFoundError
        else:
            return str(self.path.stat().st_mtime)


def _collect_nodes(
    session, path: Path, name: str, nodes: Dict[str, Union[str, Path]]
) -> Dict[str, Path]:
    """Collect nodes for a task.

    Parameters
    ----------
    session : _pytask.session.Session
        The session.
    path : Path
        The path to the task whose nodes are collected.
    name : str
        The name of the task.
    nodes : Dict[str, Union[str, Path]]
        A dictionary of nodes parsed from the ``depends_on`` or ``produces`` markers.

    Returns
    -------
    Dict[str, Path]
        A dictionary of node names and their paths.

    Raises
    ------
    NodeNotCollectedError
        If the node could not collected.

    """
    collected_nodes = {}

    for node_name, node in nodes.items():
        collected_node = session.hook.pytask_collect_node(
            session=session, path=path, node=node
        )
        if collected_node is None:
            raise NodeNotCollectedError(
                f"'{node}' cannot be parsed as a dependency or product for task "
                f"'{name}' in '{path}'."
            )
        else:
            collected_nodes[node_name] = collected_node

    return collected_nodes


def _extract_nodes_from_function_markers(function, parser):
    """Extract nodes from a marker.

    The parser is a functions which is used to document the marker with the correct
    signature. Using the function as a parser for the ``args`` and ``kwargs`` of the
    marker provides the expected error message for misspecification.

    """
    marker_name = parser.__name__
    for marker in get_marks_from_obj(function, marker_name):
        parsed = parser(*marker.args, **marker.kwargs)
        yield parsed


def _convert_objects_to_node_dictionary(objects, when):
    """Convert objects to node dictionary."""
    list_of_tuples, keep_dict = _convert_objects_to_list_of_tuples(objects, when)
    _check_that_names_are_not_used_multiple_times(list_of_tuples, when)
    nodes = _convert_nodes_to_dictionary(list_of_tuples)
    return nodes, keep_dict


def _convert_objects_to_list_of_tuples(objects, when: str):
    """Convert objects to list of tuples.

    Examples
    --------
    _convert_objects_to_list_of_tuples([{0: 0}, [4, (3, 2)], ((1, 4),))
    [(0, 0), (4,), (3, 2), (1, 4)], False

    """
    keep_dict = False

    out = []
    for obj in objects:
        if isinstance(obj, dict):
            obj = obj.items()

        if isinstance(obj, Iterable) and not isinstance(obj, str):
            keep_dict = True
            for x in obj:
                if isinstance(x, Iterable) and not isinstance(x, str):
                    tuple_x = tuple(x)
                    if len(tuple_x) in [1, 2]:
                        out.append(tuple_x)
                    else:
                        name = "Dependencies" if when == "depends_on" else "Products"
                        raise ValueError(
                            f"{name} in pytask.mark.{when} can be given as a value or "
                            "a name and a value which is 1 or 2 elements. The "
                            f"following node has {len(tuple_x)} elements: {tuple_x}."
                        )
                else:
                    out.append((x,))
        else:
            out.append((obj,))

    if len(out) > 1:
        keep_dict = False

    return out, keep_dict


def _check_that_names_are_not_used_multiple_times(list_of_tuples, when):
    """Check that names of nodes are not assigned multiple times.

    Tuples in the list have either one or two elements. The first element in the two
    element tuples is the name and cannot occur twice.

    Examples
    --------
    >>> _check_that_names_are_not_used_multiple_times(
    ...     [("a",), ("a", 1)], "depends_on"
    ... )
    >>> _check_that_names_are_not_used_multiple_times(
    ...     [("a", 0), ("a", 1)], "produces"
    ... )
    Traceback (most recent call last):
    ValueError: '@pytask.mark.produces' has nodes with the same name: {'a'}

    """
    names = [x[0] for x in list_of_tuples if len(x) == 2]
    duplicated = find_duplicates(names)

    if duplicated:
        raise ValueError(
            f"'@pytask.mark.{when}' has nodes with the same name: {duplicated}"
        )


def _convert_nodes_to_dictionary(
    list_of_tuples: List[Tuple[str]],
) -> Dict[str, Union[str, Path]]:
    """Convert nodes to dictionaries.

    Examples
    --------
    >>> _convert_nodes_to_dictionary([(0,), (1,)])
    {0: 0, 1: 1}
    >>> _convert_nodes_to_dictionary([(1, 0), (1,)])
    {1: 0, 0: 1}

    """
    nodes = {}
    counter = itertools.count()
    names = [x[0] for x in list_of_tuples if len(x) == 2]

    for tuple_ in list_of_tuples:
        if len(tuple_) == 2:
            node_name, node = tuple_
            nodes[node_name] = node

        else:
            while True:
                node_name = next(counter)
                if node_name not in names:
                    break

            nodes[node_name] = tuple_[0]

    return nodes


def create_task_name(path: Path, base_name: str):
    """Create the name of a task from a path and the task's base name.

    Examples
    --------
    >>> from pathlib import Path
    >>> create_task_name(Path("module.py"), "task_dummy")
    'module.py::task_dummy'

    """
    return path.as_posix() + "::" + base_name


def reduce_node_name(node, paths: List[Path]):
    """Reduce the node name.

    The whole name of the node - which includes the drive letter - can be very long
    when using nested folder structures in bigger projects.

    Thus, the part of the name which contains the path is replaced by the relative
    path from one path in ``session.config["paths"]`` to the node.

    """
    ancestor = find_closest_ancestor(node.path, paths)
    if ancestor is None:
        try:
            ancestor = find_common_ancestor(node.path, *paths)
        except ValueError:
            ancestor = node.path.parents[-1]

    if isinstance(node, MetaTask):
        shortened_path = relative_to(node.path, ancestor)
        raw_name = create_task_name(shortened_path, node.base_name)
        name = escape_squared_brackets(raw_name)
    elif isinstance(node, MetaNode):
        name = relative_to(node.path, ancestor).as_posix()
    else:
        raise ValueError(f"Unknown node {node} with type '{type(node)}'.")

    return name


def reduce_names_of_multiple_nodes(names, dag, paths):
    """Reduce the names of multiple nodes in the DAG."""
    return [
        reduce_node_name(dag.nodes[n].get("node") or dag.nodes[n].get("task"), paths)
        for n in names
    ]



"""Process tracebacks."""
from pathlib import Path

import _pytask
import pluggy

_PLUGGY_DIRECTORY = Path(pluggy.__file__).parent
_PYTASK_DIRECTORY = Path(_pytask.__file__).parent


def remove_traceback_from_exc_info(exc_info):
    """Remove traceback from exception."""
    return (*exc_info[:2], None)


def remove_internal_traceback_frames_from_exc_info(exc_info):
    """Remove internal traceback frames from exception info.

    If a non-internal traceback frame is found, return the traceback from the first
    occurrence downwards.

    """
    if exc_info is not None:
        filtered_traceback = _filter_internal_traceback_frames(exc_info[2])
        exc_info = (*exc_info[:2], filtered_traceback)

    return exc_info


def _is_internal_traceback_frame(frame):
    """Returns ``True`` if traceback frame belongs to internal packages.

    Internal packages are ``_pytask`` and ``pluggy``.

    """
    path = Path(frame.tb_frame.f_code.co_filename)
    return any(root in path.parents for root in [_PLUGGY_DIRECTORY, _PYTASK_DIRECTORY])


def _filter_internal_traceback_frames(frame):
    """Filter internal traceback frames from traceback.

    If the first external frame is visited, return the frame. Else return ``None``.

    """
    for frame in _yield_traceback_frames(frame):
        if frame is None or not _is_internal_traceback_frame(frame):
            break
    return frame


def _yield_traceback_frames(frame):
    """Yield traceback frames."""
    yield frame
    yield from _yield_traceback_frames(frame.tb_next)



import attr
from _pytask.enums import ExitCode


@attr.s
class Session:
    """The session of pytask."""

    config = attr.ib(factory=dict)
    """dict: A dictionary containing the configuration of the session."""
    hook = attr.ib(default=None)
    """pluggy.hooks._HookRelay: Holds all hooks collected by pytask."""
    collection_reports = attr.ib(factory=list)
    """Optional[List[pytask.report.ExecutionReport]]: Reports for collected items.

    The reports capture errors which happened while collecting tasks.

    """
    tasks = attr.ib(factory=list)
    """Optional[List[pytask.nodes.MetaTask]]: List of collected tasks."""
    deselected = attr.ib(factory=list)
    """Optional[List[pytask.nodes.MetaTask]]: list of deselected tasks."""
    resolving_dependencies_report = attr.ib(factory=list)
    """Optional[pytask.report.ResolvingDependenciesReport]: A report.

    Report when resolving dependencies failed.

    """
    execution_reports = attr.ib(factory=list)
    """Optional[List[pytask.report.ExecutionReport]]: Reports for executed tasks."""
    exit_code = attr.ib(default=ExitCode.OK)

    collection_start = attr.ib(default=None)
    collection_end = attr.ib(default=None)
    execution_start = attr.ib(default=None)
    execution_end = attr.ib(default=None)

    n_tasks_failed = attr.ib(default=0)
    """Optional[int]: Number of tests which have failed."""
    should_stop = attr.ib(default=False)
    """Optional[bool]: Indicates whether the session should be stopped."""

    @classmethod
    def from_config(cls, config):
        """Construct the class from a config."""
        return cls(config, config["pm"].hook)



"""This module contains everything related to reports."""
import attr
from _pytask.traceback import remove_internal_traceback_frames_from_exc_info


@attr.s
class CollectionReport:
    """A collection report for a task."""

    node = attr.ib(default=None)
    exc_info = attr.ib(default=None)

    @classmethod
    def from_node(cls, node):
        return cls(node=node)

    @classmethod
    def from_exception(cls, exc_info, node=None):
        exc_info = remove_internal_traceback_frames_from_exc_info(exc_info)
        return cls(exc_info=exc_info, node=node)

    @property
    def successful(self):
        return self.exc_info is None


@attr.s
class ResolvingDependenciesReport:
    """A report for an error while resolving dependencies."""

    exc_info = attr.ib()

    @classmethod
    def from_exception(cls, exc_info):
        exc_info = remove_internal_traceback_frames_from_exc_info(exc_info)
        return cls(exc_info)


@attr.s
class ExecutionReport:
    """A report for an executed task."""

    task = attr.ib()
    success = attr.ib(type=bool)
    exc_info = attr.ib(default=None)
    sections = attr.ib(factory=list)
    symbol = attr.ib(default="?")
    color = attr.ib(default=None)

    @classmethod
    def from_task_and_exception(cls, task, exc_info):
        exc_info = remove_internal_traceback_frames_from_exc_info(exc_info)
        return cls(task, False, exc_info, task._report_sections)

    @classmethod
    def from_task(cls, task):
        return cls(task, True, None, task._report_sections)



class PytaskError(Exception):
    """Base exception for pytask which should be inherited by all other exceptions."""


class NodeNotFoundError(PytaskError):
    """Exception for missing dependencies."""


class NodeNotCollectedError(PytaskError):
    """Exception for nodes which could not be collected."""


class ConfigurationError(PytaskError):
    """Exception during the configuration."""


class CollectionError(PytaskError):
    """Exception during collection."""


class ResolvingDependenciesError(PytaskError):
    """Exception during resolving dependencies."""


class ExecutionError(PytaskError):
    """Exception during execution."""



"""This module contains everything related to skipping tasks."""
from _pytask.config import hookimpl
from _pytask.dag import descending_tasks
from _pytask.enums import ColorCode
from _pytask.mark import get_specific_markers_from_task
from _pytask.mark import Mark
from _pytask.outcomes import Skipped
from _pytask.outcomes import SkippedAncestorFailed
from _pytask.outcomes import SkippedUnchanged
from _pytask.traceback import remove_traceback_from_exc_info


def skip_ancestor_failed(reason: str = "No reason provided.") -> str:
    """Function to parse information in ``@pytask.mark.skip_ancestor_failed``."""
    return reason


def skipif(condition: bool, *, reason: str) -> tuple:
    """Function to parse information in ``@pytask.mark.skipif``."""
    return condition, reason


@hookimpl
def pytask_parse_config(config):
    markers = {
        "skip": "Skip a task and all its subsequent tasks as well.",
        "skip_ancestor_failed": "Internal decorator applied to tasks whose ancestor "
        "failed.",
        "skip_unchanged": "Internal decorator applied to tasks which have already been "
        "executed and have not been changed.",
        "skipif": "Skip a task and all its subsequent tasks in case a condition is "
        "fulfilled.",
    }
    config["markers"] = {**config["markers"], **markers}


@hookimpl
def pytask_execute_task_setup(task):
    """Take a short-cut for skipped tasks during setup with an exception."""
    markers = get_specific_markers_from_task(task, "skip_unchanged")
    if markers:
        raise SkippedUnchanged

    markers = get_specific_markers_from_task(task, "skip_ancestor_failed")
    if markers:
        message = "\n".join(
            skip_ancestor_failed(*marker.args, **marker.kwargs) for marker in markers
        )
        raise SkippedAncestorFailed(message)

    markers = get_specific_markers_from_task(task, "skip")
    if markers:
        raise Skipped

    markers = get_specific_markers_from_task(task, "skipif")
    if markers:
        marker_args = [skipif(*marker.args, **marker.kwargs) for marker in markers]
        message = "\n".join(arg[1] for arg in marker_args if arg[0])
        should_skip = any(arg[0] for arg in marker_args)
        if should_skip:
            raise Skipped(message)


@hookimpl
def pytask_execute_task_process_report(session, report):
    """Process the execution reports for skipped tasks.

    This functions allows to turn skipped tasks to successful tasks.

    """
    task = report.task

    if report.exc_info:
        if isinstance(report.exc_info[1], SkippedUnchanged):
            report.success = True
            report.symbol = "s"
            report.color = ColorCode.SUCCESS

        elif isinstance(report.exc_info[1], Skipped):
            report.success = True
            report.symbol = "s"
            report.color = ColorCode.SKIPPED

            for descending_task_name in descending_tasks(task.name, session.dag):
                descending_task = session.dag.nodes[descending_task_name]["task"]
                descending_task.markers.append(
                    Mark(
                        "skip",
                        (),
                        {"reason": f"Previous task '{task.name}' was skipped."},
                    )
                )

        elif isinstance(report.exc_info[1], SkippedAncestorFailed):
            report.success = False
            report.exc_info = remove_traceback_from_exc_info(report.exc_info)
            report.symbol = "s"
            report.color = ColorCode.FAILED

    if report.exc_info and isinstance(
        report.exc_info[1], (Skipped, SkippedUnchanged, SkippedAncestorFailed)
    ):
        return True



r"""Evaluate match expressions, as used by `-k` and `-m`.

The grammar is:

| expression | expr? EOF                  |                                  |
| expr       | and_expr ('or' and_expr)*  |                                  |
| and_expr   | not_expr ('and' not_expr)* |                                  |
| not_expr   | 'not' not_expr             | '(' expr ')' or ident            |
| ident      | (\w                        | : or \+ or - or \. or \[ or \])+ |

The semantics are:

- Empty expression evaluates to False.
- ident evaluates to True of False according to a provided matcher function.
- or/and/not evaluate according to the usual boolean semantics.

"""
import ast
import enum
import re
import types
from typing import Callable
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Sequence

import attr


__all__ = ["Expression", "ParseError"]


class TokenType(enum.Enum):
    LPAREN = "left parenthesis"
    RPAREN = "right parenthesis"
    OR = "or"
    AND = "and"
    NOT = "not"
    IDENT = "identifier"
    EOF = "end of input"


@attr.s(frozen=True, slots=True)
class Token:
    type_ = attr.ib(type=TokenType)
    value = attr.ib(type=str)
    pos = attr.ib(type=int)


class ParseError(Exception):
    """The expression contains invalid syntax.

    Attributes
    ----------
    column : int
        The column in the line where the error occurred (1-based).
    message : str
        A description of the error.

    """

    def __init__(self, column: int, message: str) -> None:
        self.column = column
        self.message = message

    def __str__(self) -> str:
        return f"at column {self.column}: {self.message}"


class Scanner:
    __slots__ = ("tokens", "current")

    def __init__(self, input_: str) -> None:
        self.tokens = self.lex(input_)
        self.current = next(self.tokens)

    def lex(self, input_: str) -> Iterator[Token]:
        pos = 0
        while pos < len(input_):
            if input_[pos] in (" ", "\t"):
                pos += 1
            elif input_[pos] == "(":
                yield Token(TokenType.LPAREN, "(", pos)
                pos += 1
            elif input_[pos] == ")":
                yield Token(TokenType.RPAREN, ")", pos)
                pos += 1
            else:
                match = re.match(r"(:?\w|:|\+|-|\.|\[|\])+", input_[pos:])
                if match:
                    value = match.group(0)
                    if value == "or":
                        yield Token(TokenType.OR, value, pos)
                    elif value == "and":
                        yield Token(TokenType.AND, value, pos)
                    elif value == "not":
                        yield Token(TokenType.NOT, value, pos)
                    else:
                        yield Token(TokenType.IDENT, value, pos)
                    pos += len(value)
                else:
                    raise ParseError(
                        pos + 1,
                        f'unexpected character "{input_[pos]}"',
                    )
        yield Token(TokenType.EOF, "", pos)

    def accept(self, type_: TokenType, *, reject: bool = False) -> Optional[Token]:
        if self.current.type_ is type_:
            token = self.current
            if token.type_ is not TokenType.EOF:
                self.current = next(self.tokens)
            return token
        if reject:
            self.reject((type_,))
        return None

    def reject(self, expected: Sequence[TokenType]):
        raise ParseError(
            self.current.pos + 1,
            "expected {}; got {}".format(
                " OR ".join(type_.value for type_ in expected),
                self.current.type_.value,
            ),
        )


# True, False and None are legal match expression identifiers, but illegal as Python
# identifiers. To fix this, this prefix is added to identifiers in the conversion to
# Python AST.
IDENT_PREFIX = "$"


def expression(s: Scanner) -> ast.Expression:
    if s.accept(TokenType.EOF):
        ret: ast.expr = ast.NameConstant(False)
    else:
        ret = expr(s)
        s.accept(TokenType.EOF, reject=True)
    return ast.fix_missing_locations(ast.Expression(ret))


def expr(s: Scanner) -> ast.expr:
    ret = and_expr(s)
    while s.accept(TokenType.OR):
        rhs = and_expr(s)
        ret = ast.BoolOp(ast.Or(), [ret, rhs])
    return ret


def and_expr(s: Scanner) -> ast.expr:
    ret = not_expr(s)
    while s.accept(TokenType.AND):
        rhs = not_expr(s)
        ret = ast.BoolOp(ast.And(), [ret, rhs])
    return ret


def not_expr(s: Scanner) -> ast.expr:
    if s.accept(TokenType.NOT):
        return ast.UnaryOp(ast.Not(), not_expr(s))
    if s.accept(TokenType.LPAREN):
        ret = expr(s)
        s.accept(TokenType.RPAREN, reject=True)
        return ret
    ident = s.accept(TokenType.IDENT)
    if ident:
        return ast.Name(IDENT_PREFIX + ident.value, ast.Load())
    s.reject((TokenType.NOT, TokenType.LPAREN, TokenType.IDENT))


class MatcherAdapter(Mapping[str, bool]):
    """Adapts a matcher function to a locals mapping as required by eval()."""

    def __init__(self, matcher: Callable[[str], bool]) -> None:
        self.matcher = matcher

    def __getitem__(self, key: str) -> bool:
        return self.matcher(key[len(IDENT_PREFIX) :])

    def __iter__(self) -> Iterator[str]:
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()


class Expression:
    """A compiled match expression as used by -k and -m.

    The expression can be evaluated against different matchers.

    """

    __slots__ = ("code",)

    def __init__(self, code: types.CodeType) -> None:
        self.code = code

    @classmethod
    def compile_(cls, input_: str) -> "Expression":
        """Compile a match expression.

        Parameters
        ----------
        input_: str
            The input expression - one line.

        """
        astexpr = expression(Scanner(input_))
        code: types.CodeType = compile(
            astexpr,
            filename="<pytest match expression>",
            mode="eval",
        )
        return cls(code)

    def evaluate(self, matcher: Callable[[str], bool]) -> bool:
        """Evaluate the match expression.

        Parameters
        ----------
        matcher : Callable[[str], bool]
            Given an identifier, should return whether it matches or not. Should be
            prepared to handle arbitrary strings as input.

        Returns
        -------
        bool
            Whether the expression matches or not.

        """
        ret: bool = eval(self.code, {"__builtins__": {}}, MatcherAdapter(matcher))
        return ret



"""Add general logging capabilities."""
import platform
import sys
from typing import Any
from typing import List
from typing import Tuple

import _pytask
import pluggy
from _pytask.config import hookimpl
from _pytask.console import console


@hookimpl
def pytask_log_session_header(session):
    """Log the header of a pytask session."""
    console.rule("Start pytask session", style=None)
    console.print(
        f"Platform: {sys.platform} -- Python {platform.python_version()}, "
        f"pytask {_pytask.__version__}, pluggy {pluggy.__version__}"
    )
    console.print(f"Root: {session.config['root']}")
    if session.config["config"] is not None:
        console.print(f"Configuration: {session.config['config'].as_posix()}")

    plugin_info = session.config["pm"].list_plugin_distinfo()
    if plugin_info:
        formatted_plugins_w_versions = ", ".join(
            _format_plugin_names_and_versions(plugin_info)
        )
        console.print(f"Plugins: {formatted_plugins_w_versions}")


def _format_plugin_names_and_versions(plugininfo) -> List[str]:
    """Format name and version of loaded plugins."""
    values: List[str] = []
    for _, dist in plugininfo:
        # Gets us name and version!
        name = f"{dist.project_name}-{dist.version}"
        # Questionable convenience, but it keeps things short.
        if name.startswith("pytask-"):
            name = name[7:]
        # We decided to print python package names they can have more than one plugin.
        if name not in values:
            values.append(name)
    return values


@hookimpl
def pytask_log_session_footer(
    infos: List[Tuple[Any]], duration: float, color: str
) -> str:
    """Format the footer of the log message."""
    message = f"[{color}]"
    message += _style_infos(infos)
    message += f" in {duration}s"

    console.rule(message, style=color)


def _style_infos(infos: List[Tuple[Any]]) -> str:
    """Style infos.

    Example
    -------
    >>> m = _style_infos([(1, "a", "green"), (2, "b", "red"), (0, "c", "yellow")])
    >>> print(m)
    [green]1 a[/], [red]2 b[/]

    """
    message = []
    for value, description, color in infos:
        if value:
            message.append(f"[{color}]{value} {description}[/]")
    return ", ".join(message)



"""Implement functionality to collect tasks."""
import importlib
import inspect
import os
import sys
import time
from pathlib import Path
from typing import Generator
from typing import List

from _pytask.config import hookimpl
from _pytask.config import IS_FILE_SYSTEM_CASE_SENSITIVE
from _pytask.console import console
from _pytask.console import generate_collection_status
from _pytask.enums import ColorCode
from _pytask.exceptions import CollectionError
from _pytask.mark import has_marker
from _pytask.nodes import create_task_name
from _pytask.nodes import FilePathNode
from _pytask.nodes import PythonFunctionTask
from _pytask.nodes import reduce_node_name
from _pytask.path import find_case_sensitive_path
from _pytask.report import CollectionReport
from rich.live import Live
from rich.traceback import Traceback


@hookimpl
def pytask_collect(session):
    """Collect tasks."""
    session.collection_start = time.time()

    with Live(generate_collection_status(0), console=console, transient=True) as live:
        _collect_from_paths(session, live)

    try:
        session.hook.pytask_collect_modify_tasks(session=session, tasks=session.tasks)
    except Exception:
        report = CollectionReport.from_exception(exc_info=sys.exc_info())
        session.collection_reports.append(report)

    session.hook.pytask_collect_log(
        session=session, reports=session.collection_reports, tasks=session.tasks
    )

    return True


def _collect_from_paths(session, live=None):
    """Collect tasks from paths.

    Go through all paths, check if the path is ignored, and collect the file if not.

    """
    for path in _not_ignored_paths(session.config["paths"], session):
        reports = session.hook.pytask_collect_file_protocol(
            session=session, path=path, reports=session.collection_reports
        )
        if reports is not None:
            session.collection_reports.extend(reports)
            session.tasks.extend(i.node for i in reports if i.successful)
            if live is not None:
                live.update(generate_collection_status(len(session.tasks)))


@hookimpl
def pytask_ignore_collect(path, config):
    """Ignore a path during the collection."""
    is_ignored = any(path.match(pattern) for pattern in config["ignore"])
    return is_ignored


@hookimpl
def pytask_collect_file_protocol(session, path, reports):
    try:
        reports = session.hook.pytask_collect_file(
            session=session, path=path, reports=reports
        )
    except Exception:
        node = FilePathNode.from_path(path)
        reports = [CollectionReport.from_exception(node=node, exc_info=sys.exc_info())]

    return reports


@hookimpl
def pytask_collect_file(session, path, reports):
    """Collect a file."""
    if any(path.match(pattern) for pattern in session.config["task_files"]):
        spec = importlib.util.spec_from_file_location(path.stem, str(path))

        if spec is None:
            raise ImportError(f"Can't find module '{path.stem}' at location {path}.")

        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        collected_reports = []
        for name, obj in inspect.getmembers(mod):
            if has_marker(obj, "parametrize"):
                names_and_objects = session.hook.pytask_parametrize_task(
                    session=session, name=name, obj=obj
                )
            else:
                names_and_objects = [(name, obj)]

            for name_, obj_ in names_and_objects:
                report = session.hook.pytask_collect_task_protocol(
                    session=session, reports=reports, path=path, name=name_, obj=obj_
                )
                if report is not None:
                    collected_reports.append(report)

        return collected_reports


@hookimpl
def pytask_collect_task_protocol(session, path, name, obj):
    """Start protocol for collecting a task."""
    try:
        session.hook.pytask_collect_task_setup(
            session=session, path=path, name=name, obj=obj
        )
        task = session.hook.pytask_collect_task(
            session=session, path=path, name=name, obj=obj
        )
        if task is not None:
            session.hook.pytask_collect_task_teardown(session=session, task=task)
            return CollectionReport.from_node(task)

    except Exception:
        task = PythonFunctionTask(name, create_task_name(path, name), path, None)
        return CollectionReport.from_exception(exc_info=sys.exc_info(), node=task)


@hookimpl(trylast=True)
def pytask_collect_task(session, path, name, obj):
    """Collect a task which is a function.

    There is some discussion on how to detect functions in this `thread
    <https://stackoverflow.com/q/624926/7523785>`_. :class:`types.FunctionType` does not
    detect built-ins which is not possible anyway.

    """
    if name.startswith("task_") and callable(obj):
        return PythonFunctionTask.from_path_name_function_session(
            path, name, obj, session
        )


_TEMPLATE_ERROR = (
    "The provided path of the dependency/product in the marker is {}, but the path of "
    "the file on disk is {}. Case-sensitive file systems would raise an error.\n\n"
    "Please, align the names to ensure reproducibility on case-sensitive file systems "
    "(often Linux or macOS) or disable this error with 'check_casing_of_paths = false'."
)


@hookimpl(trylast=True)
def pytask_collect_node(session, path, node):
    """Collect a node of a task as a :class:`pytask.nodes.FilePathNode`.

    Strings are assumed to be paths. This might be a strict assumption, but since this
    hook is executed at last and possible errors will be shown, it seems reasonable and
    unproblematic.

    ``trylast=True`` might be necessary if other plugins try to parse strings themselves
    like a plugin for downloading files which depends on URLs given as strings.

    Parameters
    ----------
    session : _pytask.session.Session
        The session.
    path : Union[str, pathlib.Path]
        The path to file where the task and node are specified.
    node : Union[str, pathlib.Path]
        The value of the node which can be a str, a path or anything which cannot be
        handled by this function.

    """
    if isinstance(node, str):
        node = Path(node)
    if isinstance(node, Path):
        if not node.is_absolute():
            node = path.parent.joinpath(node)

        # ``normpath`` removes ``../`` from the path which is necessary for the casing
        # check which will fail since ``.resolves()`` also normalizes a path.
        node = Path(os.path.normpath(node))

        if (
            not IS_FILE_SYSTEM_CASE_SENSITIVE
            and session.config["check_casing_of_paths"]
            and sys.platform == "win32"
        ):
            case_sensitive_path = find_case_sensitive_path(node, "win32")
            if str(node) != str(case_sensitive_path):
                raise Exception(_TEMPLATE_ERROR.format(node, case_sensitive_path))

        return FilePathNode.from_path(node)


def _not_ignored_paths(paths: List[Path], session) -> Generator[Path, None, None]:
    """Traverse paths and yield not ignored paths.

    The paths passed by the user can either point to files or directories. For
    directories, all subsequent files and folders are considered, but one level after
    another, so that files of ignored folders are not checked.

    Parameters
    ----------
    paths : List[pathlib.Path]
        List of paths from which tasks are collected.
    session : _pytask.session.Session
        The session.

    Yields
    ------
    path : pathlib.Path
        A path which is not ignored.

    """
    for path in paths:
        if not session.hook.pytask_ignore_collect(path=path, config=session.config):
            if path.is_dir():
                files_in_dir = path.iterdir()
                yield from _not_ignored_paths(files_in_dir, session)
            else:
                yield path


@hookimpl
def pytask_collect_log(session, reports, tasks):
    """Log collection."""
    session.collection_end = time.time()

    message = f"Collected {len(tasks)} task{'' if len(tasks) == 1 else 's'}."

    n_deselected = len(session.deselected)
    if n_deselected:
        message += f" Deselected {n_deselected} task{'' if n_deselected == 1 else 's'}."
    console.print(message)

    failed_reports = [i for i in reports if not i.successful]
    if failed_reports:
        console.print()
        console.rule(
            f"[{ColorCode.FAILED}]Failures during collection", style=ColorCode.FAILED
        )

        for report in failed_reports:
            if report.node is None:
                header = "Error"
            else:
                short_name = reduce_node_name(report.node, session.config["paths"])
                header = f"Could not collect {short_name}"

            console.rule(f"[{ColorCode.FAILED}]{header}", style=ColorCode.FAILED)

            console.print()

            console.print(
                Traceback.from_exception(
                    *report.exc_info, show_locals=session.config["show_locals"]
                )
            )

            console.print()

        session.hook.pytask_log_session_footer(
            session=session,
            infos=[
                (len(tasks), "collected", ColorCode.SUCCESS),
                (len(failed_reports), "failed", ColorCode.FAILED),
                (n_deselected, "deselected", ColorCode.NEUTRAL),
            ],
            duration=round(session.collection_end - session.collection_start, 2),
            color=ColorCode.FAILED if len(failed_reports) else ColorCode.SUCCESS,
        )

        raise CollectionError



"""This module holds the plugin manager."""
import pluggy
from _pytask import hookspecs


def get_plugin_manager():
    """Get the plugin manager."""
    pm = pluggy.PluginManager("pytask")
    pm.add_hookspecs(hookspecs)
    pm.load_setuptools_entrypoints("pytask")

    return pm



"""Implement the build command."""
import sys

import click
from _pytask.config import hookimpl
from _pytask.console import console
from _pytask.enums import ExitCode
from _pytask.exceptions import CollectionError
from _pytask.exceptions import ConfigurationError
from _pytask.exceptions import ExecutionError
from _pytask.exceptions import ResolvingDependenciesError
from _pytask.pluginmanager import get_plugin_manager
from _pytask.session import Session


@hookimpl(tryfirst=True)
def pytask_extend_command_line_interface(cli):
    """Extend the command line interface."""
    cli.add_command(build)


def main(config_from_cli):
    """Run pytask.

    This is the main command to run pytask which usually receives kwargs from the
    command line interface. It can also be used to run pytask interactively. Pass
    configuration in a dictionary.

    Parameters
    ----------
    config_from_cli : dict
        A dictionary with options passed to pytask. In general, this dictionary holds
        the information passed via the command line interface.

    Returns
    -------
    session : pytask.session.Session
        The session captures all the information of the current run.

    """
    try:
        pm = get_plugin_manager()
        from _pytask import cli

        pm.register(cli)
        pm.hook.pytask_add_hooks(pm=pm)

        config = pm.hook.pytask_configure(pm=pm, config_from_cli=config_from_cli)

        session = Session.from_config(config)

    except (ConfigurationError, Exception):
        console.print_exception()
        session = Session({}, None)
        session.exit_code = ExitCode.CONFIGURATION_FAILED

    else:
        try:
            session.hook.pytask_log_session_header(session=session)
            session.hook.pytask_collect(session=session)
            session.hook.pytask_resolve_dependencies(session=session)
            session.hook.pytask_execute(session=session)

        except CollectionError:
            session.exit_code = ExitCode.COLLECTION_FAILED

        except ResolvingDependenciesError:
            session.exit_code = ExitCode.RESOLVING_DEPENDENCIES_FAILED

        except ExecutionError:
            session.exit_code = ExitCode.FAILED

        except Exception:
            console.print_exception()
            session.exit_code = ExitCode.FAILED

        session.hook.pytask_unconfigure(session=session)

    return session


@click.command()
@click.option(
    "--debug-pytask",
    is_flag=True,
    default=None,
    help="Trace all function calls in the plugin framework.  [default: False]",
)
@click.option(
    "-x",
    "--stop-after-first-failure",
    is_flag=True,
    default=None,
    help="Stop after the first failure.",
)
@click.option("--max-failures", default=None, help="Stop after some failures.")
def build(**config_from_cli):
    """Collect and execute tasks and report the results.

    This is the default command of pytask which searches given paths or the current
    working directory for tasks to execute them. A report informs you on the results.

    """
    config_from_cli["command"] = "build"
    session = main(config_from_cli)
    sys.exit(session.exit_code)



"""Enumerations for pytask."""
import enum


class ExitCode(enum.IntEnum):
    """Exit codes for pytask."""

    OK = 0
    """Tasks were executed successfully."""

    FAILED = 1
    """Failed while executing tasks."""

    CONFIGURATION_FAILED = 2

    COLLECTION_FAILED = 3
    """Failed while collecting tasks."""

    RESOLVING_DEPENDENCIES_FAILED = 4
    """Failed while resolving dependencies."""


class ColorCode(str, enum.Enum):
    """Color codes for pytask."""

    SUCCESS = "green"

    FAILED = "red"

    SKIPPED = "yellow"

    NEUTRAL = None

    def __str__(self) -> str:
        """Allows to use a color code inside f-strings without the need for .value."""
        return str.__str__(self)



"""Implement the ability for tasks to persist."""
from _pytask.config import hookimpl
from _pytask.dag import node_and_neighbors
from _pytask.enums import ColorCode
from _pytask.exceptions import NodeNotFoundError
from _pytask.mark import get_specific_markers_from_task
from _pytask.outcomes import Persisted


@hookimpl
def pytask_parse_config(config):
    """Add the marker to the configuration."""
    config["markers"]["persist"] = (
        "Prevent execution of a task if all products exist and even if something has "
        "changed (dependencies, source file, products). This decorator might be useful "
        "for expensive tasks where only the formatting of the file has changed. The "
        "state of the files which have changed will also be remembered and another run "
        "will skip the task with success."
    )


@hookimpl
def pytask_execute_task_setup(session, task):
    """Exit persisting tasks early.

    The decorator needs to be set and all nodes need to exist.

    """
    if get_specific_markers_from_task(task, "persist"):
        try:
            for name in node_and_neighbors(session.dag, task.name):
                node = (
                    session.dag.nodes[name].get("task")
                    or session.dag.nodes[name]["node"]
                )
                node.state()
        except NodeNotFoundError:
            all_nodes_exist = False
        else:
            all_nodes_exist = True

        if all_nodes_exist:
            raise Persisted


@hookimpl
def pytask_execute_task_process_report(report):
    """Set task status to success.

    Do not return ``True`` so that states will be updated in database.

    """
    if report.exc_info and isinstance(report.exc_info[1], Persisted):
        report.success = True
        report.symbol = "p"
        report.color = ColorCode.SUCCESS



"""Capture stdout and stderr during collection and execution.

This module implements the :class:`CaptureManager` plugin which allows for capturing in
three ways.

- fd (file descriptor) level capturing (default): All writes going to the operating
  system file descriptors 1 and 2 will be captured.
- sys level capturing: Only writes to Python files ``sys.stdout`` and ``sys.stderr``
  will be captured. No capturing of writes to file descriptors is performed.
- tee-sys capturing: Python writes to ``sys.stdout`` and ``sys.stderr`` will be
  captured, however the writes will also be passed-through to the actual ``sys.stdout``
  and ``sys.stderr``.


References
----------

- `Blog post on redirecting and file descriptors
  <https://eli.thegreenplace.net/2015/redirecting-all-kinds-of-stdout-in-python>`_.
- `The capture module in pytest
  <https://github.com/pytest-dev/pytest/blob/master/src/_pytest/capture.py>`_.
- `The debugging module in pytest
  <https://github.com/pytest-dev/pytest/blob/master/src/_pytest/debugging.py>`_.

"""
import contextlib
import functools
import io
import os
import sys
from tempfile import TemporaryFile
from typing import AnyStr
from typing import Generator
from typing import Generic
from typing import Iterator
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import click
from _pytask.config import hookimpl
from _pytask.nodes import MetaTask
from _pytask.shared import get_first_non_none_value

if TYPE_CHECKING:
    from typing_extensions import Literal

    _CaptureMethod = Literal["fd", "sys", "no", "tee-sys"]

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import final as final
    else:
        from typing_extensions import final as final
elif sys.version_info >= (3, 8):
    from typing import final as final
else:

    def final(f):
        return f


@hookimpl
def pytask_extend_command_line_interface(cli):
    """Add CLI options for capturing output."""
    additional_parameters = [
        click.Option(
            ["--capture"],
            type=click.Choice(["fd", "no", "sys", "tee-sys"]),
            help="Per task capturing method.  [default: fd]",
        ),
        click.Option(["-s"], is_flag=True, help="Shortcut for --capture=no."),
        click.Option(
            ["--show-capture"],
            type=click.Choice(["no", "stdout", "stderr", "all"]),
            help=(
                "Choose which captured output should be shown for failed tasks.  "
                "[default: all]"
            ),
        ),
    ]
    cli.commands["build"].params.extend(additional_parameters)


@hookimpl
def pytask_parse_config(config, config_from_cli, config_from_file):
    """Parse configuration.

    Note that, ``-s`` is a shortcut for ``--capture=no``.

    """
    if config_from_cli.get("s"):
        config["capture"] = "no"
    else:
        config["capture"] = get_first_non_none_value(
            config_from_cli,
            config_from_file,
            key="capture",
            default="fd",
            callback=_capture_callback,
        )

    config["show_capture"] = get_first_non_none_value(
        config_from_cli,
        config_from_file,
        key="show_capture",
        default="all",
        callback=_show_capture_callback,
    )


@hookimpl
def pytask_post_parse(config):
    """Initialize the CaptureManager."""
    if config["capture"] == "fd":
        _py36_windowsconsoleio_workaround(sys.stdout)
    _colorama_workaround()
    _readline_workaround()

    pluginmanager = config["pm"]
    capman = CaptureManager(config["capture"])
    pluginmanager.register(capman, "capturemanager")
    capman.stop_capturing()
    capman.start_capturing()
    capman.suspend()


def _capture_callback(x):
    """Validate the passed options for capturing output."""
    if x in [None, "None", "none"]:
        x = None
    elif x in ["fd", "no", "sys", "tee-sys"]:
        pass
    else:
        raise ValueError("'capture' can only be one of ['fd', 'no', 'sys', 'tee-sys'].")

    return x


def _show_capture_callback(x):
    """Validate the passed options for showing captured output."""
    if x in [None, "None", "none"]:
        x = None
    elif x in ["no", "stdout", "stderr", "all"]:
        pass
    else:
        raise ValueError(
            "'show_capture' must be one of ['no', 'stdout', 'stderr', 'all']."
        )

    return x


# Copied from pytest with slightly modified docstrings.


def _colorama_workaround() -> None:
    """Ensure colorama is imported so that it attaches to the correct stdio
    handles on Windows.

    colorama uses the terminal on import time. So if something does the
    first import of colorama while I/O capture is active, colorama will
    fail in various ways.

    """
    if sys.platform.startswith("win32"):
        try:
            import colorama  # noqa: F401
        except ImportError:
            pass


def _readline_workaround() -> None:
    """Ensure readline is imported so that it attaches to the correct stdio handles on
    Windows.

    Pdb uses readline support where available -- when not running from the Python
    prompt, the readline module is not imported until running the pdb REPL.  If running
    pytest with the ``--pdb`` option this means the readline module is not imported
    until after I/O capture has been started.

    This is a problem for pyreadline, which is often used to implement readline support
    on Windows, as it does not attach to the correct handles for stdout and/or stdin if
    they have been redirected by the FDCapture mechanism.  This workaround ensures that
    readline is imported before I/O capture is setup so that it can attach to the actual
    stdin/out for the console.

    See https://github.com/pytest-dev/pytest/pull/1281.

    """
    if sys.platform.startswith("win32"):
        try:
            import readline  # noqa: F401
        except ImportError:
            pass


def _py36_windowsconsoleio_workaround(stream: TextIO) -> None:
    """Workaround for Windows Unicode console handling on Python>=3.6.

    Python 3.6 implemented Unicode console handling for Windows. This works by
    reading/writing to the raw console handle using ``{Read,Write}ConsoleW``.

    The problem is that we are going to ``dup2`` over the stdio file descriptors when
    doing ``FDCapture`` and this will ``CloseHandle`` the handles used by Python to
    write to the console. Though there is still some weirdness and the console handle
    seems to only be closed randomly and not on the first call to ``CloseHandle``, or
    maybe it gets reopened with the same handle value when we suspend capturing.

    The workaround in this case will reopen stdio with a different fd which also means a
    different handle by replicating the logic in
    "Py_lifecycle.c:initstdio/create_stdio".

    Parameters
    ---------
    stream
        In practice ``sys.stdout`` or ``sys.stderr``, but given here as parameter for
        unit testing purposes.

    See https://github.com/pytest-dev/py/issues/103.

    """
    if not sys.platform.startswith("win32") or hasattr(sys, "pypy_version_info"):
        return

    # Bail out if ``stream`` doesn't seem like a proper ``io`` stream (#2666).
    if not hasattr(stream, "buffer"):  # type: ignore[unreachable]
        return

    buffered = hasattr(stream.buffer, "raw")
    raw_stdout = stream.buffer.raw if buffered else stream.buffer

    if not isinstance(raw_stdout, io._WindowsConsoleIO):  # type: ignore[attr-defined]
        return

    def _reopen_stdio(f, mode):
        if not buffered and mode[0] == "w":
            buffering = 0
        else:
            buffering = -1

        return io.TextIOWrapper(
            open(os.dup(f.fileno()), mode, buffering),  # type: ignore[arg-type]
            f.encoding,
            f.errors,
            f.newlines,
            f.line_buffering,
        )

    sys.stdin = _reopen_stdio(sys.stdin, "rb")
    sys.stdout = _reopen_stdio(sys.stdout, "wb")
    sys.stderr = _reopen_stdio(sys.stderr, "wb")


# IO Helpers.


class EncodedFile(io.TextIOWrapper):
    __slots__ = ()

    @property
    def name(self) -> str:
        # Ensure that file.name is a string. Workaround for a Python bug fixed in
        # >=3.7.4: https://bugs.python.org/issue36015
        return repr(self.buffer)

    @property
    def mode(self) -> str:
        # TextIOWrapper doesn't expose a mode, but at least some of our tests check it.
        return self.buffer.mode.replace("b", "")


class CaptureIO(io.TextIOWrapper):
    def __init__(self) -> None:
        super().__init__(io.BytesIO(), encoding="UTF-8", newline="", write_through=True)

    def getvalue(self) -> str:
        assert isinstance(self.buffer, io.BytesIO)
        return self.buffer.getvalue().decode("UTF-8")


class TeeCaptureIO(CaptureIO):
    def __init__(self, other: TextIO) -> None:
        self._other = other
        super().__init__()

    def write(self, s: str) -> int:
        super().write(s)
        return self._other.write(s)


class DontReadFromInput:
    """Class to disable reading from stdin while capturing is activated."""

    encoding = None

    def read(self, *_args):  # noqa: U101
        raise OSError(
            "pytest: reading from stdin while output is captured! Consider using `-s`."
        )

    readline = read
    readlines = read
    __next__ = read

    def __iter__(self):
        return self

    def fileno(self) -> int:
        raise io.UnsupportedOperation("redirected stdin is pseudofile, has no fileno()")

    def isatty(self) -> bool:
        return False

    def close(self) -> None:
        pass

    @property
    def buffer(self):
        return self


# Capture classes.


patchsysdict = {0: "stdin", 1: "stdout", 2: "stderr"}
"""Dict[int, str]: Map file descriptors to their names."""


class NoCapture:
    """Dummy class when capturing is disabled."""

    EMPTY_BUFFER = None
    __init__ = start = done = suspend = resume = lambda *_args: None  # noqa: U101


class SysCaptureBinary:
    """Capture IO to/from Python's buffer for stdin, stdout, and stderr.

    Instead of :class:`SysCapture`, this class produces bytes instead of text.

    """

    EMPTY_BUFFER = b""

    def __init__(self, fd: int, tmpfile=None, *, tee: bool = False) -> None:
        name = patchsysdict[fd]
        self._old = getattr(sys, name)
        self.name = name
        if tmpfile is None:
            if name == "stdin":
                tmpfile = DontReadFromInput()
            else:
                tmpfile = CaptureIO() if not tee else TeeCaptureIO(self._old)
        self.tmpfile = tmpfile
        self._state = "initialized"

    def repr(self, class_name: str) -> str:  # noqa: A003
        return "<{} {} _old={} _state={!r} tmpfile={!r}>".format(
            class_name,
            self.name,
            hasattr(self, "_old") and repr(self._old) or "<UNSET>",
            self._state,
            self.tmpfile,
        )

    def __repr__(self) -> str:
        return "<{} {} _old={} _state={!r} tmpfile={!r}>".format(
            self.__class__.__name__,
            self.name,
            hasattr(self, "_old") and repr(self._old) or "<UNSET>",
            self._state,
            self.tmpfile,
        )

    def _assert_state(self, op: str, states: Tuple[str, ...]) -> None:
        assert (
            self._state in states
        ), "cannot {} in state {!r}: expected one of {}".format(
            op, self._state, ", ".join(states)
        )

    def start(self) -> None:
        self._assert_state("start", ("initialized",))
        setattr(sys, self.name, self.tmpfile)
        self._state = "started"

    def snap(self):
        self._assert_state("snap", ("started", "suspended"))
        self.tmpfile.seek(0)
        res = self.tmpfile.buffer.read()
        self.tmpfile.seek(0)
        self.tmpfile.truncate()
        return res

    def done(self) -> None:
        self._assert_state("done", ("initialized", "started", "suspended", "done"))
        if self._state == "done":
            return
        setattr(sys, self.name, self._old)
        del self._old
        self.tmpfile.close()
        self._state = "done"

    def suspend(self) -> None:
        self._assert_state("suspend", ("started", "suspended"))
        setattr(sys, self.name, self._old)
        self._state = "suspended"

    def resume(self) -> None:
        self._assert_state("resume", ("started", "suspended"))
        if self._state == "started":
            return
        setattr(sys, self.name, self.tmpfile)
        self._state = "started"

    def writeorg(self, data) -> None:
        self._assert_state("writeorg", ("started", "suspended"))
        self._old.flush()
        self._old.buffer.write(data)
        self._old.buffer.flush()


class SysCapture(SysCaptureBinary):
    """Capture IO to/from Python's buffer for stdin, stdout, and stderr.

    Instead of :class:`SysCaptureBinary`, this class produces text instead of bytes.

    """

    EMPTY_BUFFER = ""  # type: ignore[assignment]

    def snap(self):
        res = self.tmpfile.getvalue()
        self.tmpfile.seek(0)
        self.tmpfile.truncate()
        return res

    def writeorg(self, data):
        self._assert_state("writeorg", ("started", "suspended"))
        self._old.write(data)
        self._old.flush()


class FDCaptureBinary:
    """Capture IO to/from a given OS-level file descriptor.

    snap() produces `bytes`.

    """

    EMPTY_BUFFER = b""

    def __init__(self, targetfd: int) -> None:
        self.targetfd = targetfd

        try:
            os.fstat(targetfd)
        except OSError:
            # FD capturing is conceptually simple -- create a temporary file, redirect
            # the FD to it, redirect back when done. But when the target FD is invalid
            # it throws a wrench into this lovely scheme.

            # Tests themselves shouldn't care if the FD is valid, FD capturing should
            # work regardless of external circumstances. So falling back to just sys
            # capturing is not a good option.

            # Further complications are the need to support suspend() and the
            # possibility of FD reuse (e.g. the tmpfile getting the very same target
            # FD). The following approach is robust, I believe.
            self.targetfd_invalid: Optional[int] = os.open(os.devnull, os.O_RDWR)
            os.dup2(self.targetfd_invalid, targetfd)
        else:
            self.targetfd_invalid = None
        self.targetfd_save = os.dup(targetfd)

        if targetfd == 0:
            self.tmpfile = open(os.devnull)
            self.syscapture = SysCapture(targetfd)
        else:
            self.tmpfile = EncodedFile(
                TemporaryFile(buffering=0),  # type: ignore[arg-type]
                encoding="utf-8",
                errors="replace",
                newline="",
                write_through=True,
            )
            if targetfd in patchsysdict:
                self.syscapture = SysCapture(targetfd, self.tmpfile)
            else:
                self.syscapture = NoCapture()

        self._state = "initialized"

    def __repr__(self) -> str:
        return "<{} {} oldfd={} _state={!r} tmpfile={!r}>".format(
            self.__class__.__name__,
            self.targetfd,
            self.targetfd_save,
            self._state,
            self.tmpfile,
        )

    def _assert_state(self, op: str, states: Tuple[str, ...]) -> None:
        assert (
            self._state in states
        ), "cannot {} in state {!r}: expected one of {}".format(
            op, self._state, ", ".join(states)
        )

    def start(self) -> None:
        """Start capturing on targetfd using memorized tmpfile."""
        self._assert_state("start", ("initialized",))
        os.dup2(self.tmpfile.fileno(), self.targetfd)
        self.syscapture.start()
        self._state = "started"

    def snap(self):
        self._assert_state("snap", ("started", "suspended"))
        self.tmpfile.seek(0)
        res = self.tmpfile.buffer.read()
        self.tmpfile.seek(0)
        self.tmpfile.truncate()
        return res

    def done(self) -> None:
        """Stop capturing, restore streams, return original capture file,
        seeked to position zero."""
        self._assert_state("done", ("initialized", "started", "suspended", "done"))
        if self._state == "done":
            return
        os.dup2(self.targetfd_save, self.targetfd)
        os.close(self.targetfd_save)
        if self.targetfd_invalid is not None:
            if self.targetfd_invalid != self.targetfd:
                os.close(self.targetfd)
            os.close(self.targetfd_invalid)
        self.syscapture.done()
        self.tmpfile.close()
        self._state = "done"

    def suspend(self) -> None:
        self._assert_state("suspend", ("started", "suspended"))
        if self._state == "suspended":
            return
        self.syscapture.suspend()
        os.dup2(self.targetfd_save, self.targetfd)
        self._state = "suspended"

    def resume(self) -> None:
        self._assert_state("resume", ("started", "suspended"))
        if self._state == "started":
            return
        self.syscapture.resume()
        os.dup2(self.tmpfile.fileno(), self.targetfd)
        self._state = "started"

    def writeorg(self, data):
        """Write to original file descriptor."""
        self._assert_state("writeorg", ("started", "suspended"))
        os.write(self.targetfd_save, data)


class FDCapture(FDCaptureBinary):
    """Capture IO to/from a given OS-level file descriptor.

    snap() produces text.

    """

    # Ignore type because it doesn't match the type in the superclass (bytes).
    EMPTY_BUFFER = ""  # type: ignore

    def snap(self):
        self._assert_state("snap", ("started", "suspended"))
        self.tmpfile.seek(0)
        res = self.tmpfile.read()
        self.tmpfile.seek(0)
        self.tmpfile.truncate()
        return res

    def writeorg(self, data):
        """Write to original file descriptor."""
        super().writeorg(data.encode("utf-8"))


# MultiCapture


@final
@functools.total_ordering
class CaptureResult(Generic[AnyStr]):
    """The result of :meth:`MultiCapture.readouterr` which wraps stdout and stderr.

    This class was a namedtuple, but due to mypy limitation [0]_ it could not be made
    generic, so was replaced by a regular class which tries to emulate the pertinent
    parts of a namedtuple. If the mypy limitation is ever lifted, can make it a
    namedtuple again.

    .. [0] https://github.com/python/mypy/issues/685

    """

    # Can't use slots in Python<3.5.3 due to https://bugs.python.org/issue31272
    if sys.version_info >= (3, 5, 3):
        __slots__ = ("out", "err")

    def __init__(self, out: AnyStr, err: AnyStr) -> None:
        self.out: AnyStr = out
        self.err: AnyStr = err

    def __len__(self) -> int:
        return 2

    def __iter__(self) -> Iterator[AnyStr]:
        return iter((self.out, self.err))

    def __getitem__(self, item: int) -> AnyStr:
        return tuple(self)[item]

    def _replace(
        self, *, out: Optional[AnyStr] = None, err: Optional[AnyStr] = None
    ) -> "CaptureResult[AnyStr]":
        return CaptureResult(
            out=self.out if out is None else out, err=self.err if err is None else err
        )

    def count(self, value: AnyStr) -> int:
        return tuple(self).count(value)

    def index(self, value) -> int:
        return tuple(self).index(value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (CaptureResult, tuple)):
            return NotImplemented
        return tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, (CaptureResult, tuple)):
            return NotImplemented
        return tuple(self) < tuple(other)

    def __repr__(self) -> str:
        return f"CaptureResult(out={self.out!r}, err={self.err!r})"


class MultiCapture(Generic[AnyStr]):
    """The class which manages the buffers connected to each stream.

    The class is instantiated with buffers for ``stdin``, ``stdout`` and ``stderr``.
    Then, the instance provides convenient methods to control all buffers at once, like
    start and stop capturing and reading the ``stdout`` and ``stderr``.

    """

    _state = None
    _in_suspended = False

    def __init__(self, in_, out, err) -> None:
        self.in_ = in_
        self.out = out
        self.err = err

    def __repr__(self) -> str:
        return (
            "<MultiCapture out={!r} err={!r} in_={!r} _state={!r} "
            "_in_suspended={!r}>"
        ).format(
            self.out,
            self.err,
            self.in_,
            self._state,
            self._in_suspended,
        )

    def start_capturing(self) -> None:
        self._state = "started"
        if self.in_:
            self.in_.start()
        if self.out:
            self.out.start()
        if self.err:
            self.err.start()

    def pop_outerr_to_orig(self) -> Tuple[AnyStr, AnyStr]:
        """Pop current snapshot out/err capture and flush to orig streams."""
        out, err = self.readouterr()
        if out:
            self.out.writeorg(out)
        if err:
            self.err.writeorg(err)
        return out, err

    def suspend_capturing(self, in_: bool = False) -> None:
        self._state = "suspended"
        if self.out:
            self.out.suspend()
        if self.err:
            self.err.suspend()
        if in_ and self.in_:
            self.in_.suspend()
            self._in_suspended = True

    def resume_capturing(self) -> None:
        self._state = "started"
        if self.out:
            self.out.resume()
        if self.err:
            self.err.resume()
        if self._in_suspended:
            self.in_.resume()
            self._in_suspended = False

    def stop_capturing(self) -> None:
        """Stop capturing and reset capturing streams."""
        if self._state == "stopped":
            raise ValueError("was already stopped")
        self._state = "stopped"
        if self.out:
            self.out.done()
        if self.err:
            self.err.done()
        if self.in_:
            self.in_.done()

    def is_started(self) -> bool:
        """Whether actively capturing -- not suspended or stopped."""
        return self._state == "started"

    def readouterr(self) -> CaptureResult[AnyStr]:
        if self.out:
            out = self.out.snap()
        else:
            out = ""
        if self.err:
            err = self.err.snap()
        else:
            err = ""
        return CaptureResult(out, err)


def _get_multicapture(method: "_CaptureMethod") -> MultiCapture[str]:
    """Set up the MultiCapture class with the passed method.

    For each valid method, the function instantiates the :class:`MultiCapture` class
    with the specified buffers for ``stdin``, ``stdout``, and ``stderr``.

    """
    if method == "fd":
        return MultiCapture(in_=FDCapture(0), out=FDCapture(1), err=FDCapture(2))
    elif method == "sys":
        return MultiCapture(in_=SysCapture(0), out=SysCapture(1), err=SysCapture(2))
    elif method == "no":
        return MultiCapture(in_=None, out=None, err=None)
    elif method == "tee-sys":
        return MultiCapture(
            in_=None, out=SysCapture(1, tee=True), err=SysCapture(2, tee=True)
        )
    raise ValueError(f"unknown capturing method: {method!r}")


# Own implementation of the CaptureManager.


class CaptureManager:
    """The capture plugin.

    This class is the capture plugin which implements some hooks and provides an
    interface around :func:`_get_multicapture` and :class:`MultiCapture` adjusted to
    pytask.

    The class manages that the appropriate capture method is enabled/disabled during the
    execution phase (setup, call, teardown). After each of those points, the captured
    output is obtained and attached to the execution report.

    """

    def __init__(self, method: "_CaptureMethod") -> None:
        self._method = method
        self._capturing: Optional[MultiCapture[str]] = None

    def __repr__(self) -> str:
        return ("<CaptureManager _method={!r} _capturing={!r}>").format(
            self._method, self._capturing
        )

    def is_capturing(self) -> Union[str, bool]:
        return self._method != "no"

    def start_capturing(self) -> None:
        assert self._capturing is None
        self._capturing = _get_multicapture(self._method)
        self._capturing.start_capturing()

    def stop_capturing(self) -> None:
        if self._capturing is not None:
            self._capturing.pop_outerr_to_orig()
            self._capturing.stop_capturing()
            self._capturing = None

    def resume(self) -> None:
        # During teardown of the python process, and on rare occasions, capture
        # attributes can be `None` while trying to resume global capture.
        if self._capturing is not None:
            self._capturing.resume_capturing()

    def suspend(self, in_: bool = False) -> None:
        if self._capturing is not None:
            self._capturing.suspend_capturing(in_=in_)

    def read(self) -> CaptureResult[str]:
        assert self._capturing is not None
        return self._capturing.readouterr()

    # Helper context managers

    @contextlib.contextmanager
    def task_capture(self, when: str, task: MetaTask) -> Generator[None, None, None]:
        """Pipe captured stdout and stderr into report sections."""
        self.resume()

        try:
            yield
        finally:
            self.suspend(in_=False)

        out, err = self.read()
        task.add_report_section(when, "stdout", out)
        task.add_report_section(when, "stderr", err)

    # Hooks

    @hookimpl(hookwrapper=True)
    def pytask_execute_task_setup(self, task: MetaTask) -> Generator[None, None, None]:
        """Capture output during setup."""
        with self.task_capture("setup", task):
            yield

    @hookimpl(hookwrapper=True)
    def pytask_execute_task(self, task: MetaTask) -> Generator[None, None, None]:
        """Capture output during execution."""
        with self.task_capture("call", task):
            yield

    @hookimpl(hookwrapper=True)
    def pytask_execute_task_teardown(
        self, task: MetaTask
    ) -> Generator[None, None, None]:
        """Capture output during teardown."""
        with self.task_capture("teardown", task):
            yield
